"""tidy3d mode solver.

tidy3d has a powerful open source mode solver.

tidy3d can:

- compute bend modes.
- compute mode overlaps.

"""

from __future__ import annotations

import itertools as it
import pathlib
import subprocess
import sys
import time
from pathlib import Path
from types import SimpleNamespace
from typing import Dict, List, Optional, Tuple, Union
import hashlib

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors
from pydantic import BaseModel, Extra
import pydantic
from scipy.constants import c as SPEED_OF_LIGHT
from scipy.interpolate import griddata
from tidy3d.plugins.mode.solver import compute_modes
from tqdm.auto import tqdm
from typing_extensions import Literal

from gdsfactory.serialization import clean_value_name
from gdsfactory.config import logger
from gdsfactory.pdk import MaterialSpec, get_material_index, get_modes_path
from gdsfactory.simulation.gtidy3d.materials import si, sin, sio2
from gdsfactory.typings import PathType

nm = 1e-3


def plot(
    X,
    Y,
    n,
    mode=None,
    num_levels: int = 8,
    n_cmap=None,
    mode_cmap=None,
    axes="xy",
    title=None,
    normalize_mode: bool = False,
    plot_power: bool = True,
) -> None:
    """Plot waveguide index or mode in matplotlib.

    Args:
        X: x array.
        Y: y array.
        n: refractive index.
        mode: mode number.
        num_levels: for plot.
        n_cmap: refractive index color map.
        mode_cmap: mode color map.
        axes: "xy".
        title: for the plot.
        normalize_mode: divide by maximum value.

    """
    x, y = axes
    if n_cmap is None:
        n_cmap = colors.LinearSegmentedColormap.from_list(
            name="custom", colors=["#ffffff", "#c1d9ed"]
        )
    if mode_cmap is None:
        mode_cmap = "inferno"
    if mode is not None:
        mode = np.abs(mode) if plot_power else mode.real
        if normalize_mode:
            mode = mode / mode.max()
        plt.contour(
            X,
            Y,
            mode,
            cmap=mode_cmap,
            levels=np.linspace(mode.min(), mode.max(), num_levels),
        )
        plt.colorbar(label="mode")
    n = np.real(n)
    plt.pcolormesh(X, Y, n, cmap=n_cmap)
    plt.colorbar(label="n")
    plt.xlabel(x)
    plt.ylabel(y)
    plt.grid(True, alpha=0.4)
    plt.axis("scaled")
    if title is not None:
        plt.title(title)


def _mesh_1d(x_min, x_max, mesh_x):
    if isinstance(mesh_x, int):
        x = np.linspace(x_min, x_max, mesh_x + 1)
        dx = x[1] - x[0]
    elif isinstance(mesh_x, float):
        dx = mesh_x
        x = np.arange(x_min, x_max + 0.999 * dx, dx)
        mesh_x = x.shape[0]
    else:
        raise ValueError("Invalid 'mesh_x'")
    return x, mesh_x, dx


def create_mesh(x_min, y_min, x_max, y_max, mesh_x, mesh_y):
    x, mesh_x, dx = _mesh_1d(x_min, x_max, mesh_x)
    y, mesh_y, dy = _mesh_1d(y_min, y_max, mesh_y)

    Yx, Xx = np.meshgrid(y[:-1], x[:-1] + dx / 2)
    Yy, Xy = np.meshgrid(y[:-1] + dy / 2, x[:-1])
    Yz, Xz = np.meshgrid(y[:-1], x[:-1])
    return x, y, Xx, Yx, Xy, Yy, Xz, Yz


SETTINGS = [
    "wavelength",
    "wg_width",
    "wg_thickness",
    "ncore",
    "nclad",
    "slab_thickness",
    "t_box",
    "t_clad",
    "xmargin",
    "resolution",
    "nmodes",
    "bend_radius",
]

SETTINGS_COUPLER = set(SETTINGS + ["wg_width1", "wg_width2", "gap"])
Precision = Literal["single", "double"]
FilterPol = Literal["te", "tm"]


class Waveguide(BaseModel):
    """Waveguide Model.

    Parameters:
        wavelength: (um).
        wg_width: waveguide width.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index material spec. Can be:
            string: material name.
            float: refractive index.
            float, float: refractive index real and imaginary part.
            function: function of wavelength.
        nclad: cladding refractive index material spec.
        dn_dict: unstructured mesh array with columns field "x", "y", "dn" of local index perturbations to be interpolated.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um. Can be a single number or tuple (x, y).
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
        cache: True uses file cache from PDK.modes_path. False skips cache.
        precision: single or double.
        filter_pol: te, tm or None.
        loss_model: whether to include a scattering loss region at the interfaces. Default is False
        sidewall_sigma: size of the region to append to the sidewalls in the loss model. Default is 10 nm.
        sidewall_k: imaginary index addition for the sidewall loss region. Default is 0 (no extra loss).
        top_sigma: size of the loss region to append to the top surfaces in the loss model. Default is 10 nm.
        top_k: imaginary index addition for the top surfaces loss region. Default is 0 (no extra loss).

    ::

          _________________________


                  width     xmargin
              <----------> <------>
               ___________   _ _ _
              |           |       |
          ____|  ncore    |_______
                                  | wg_thickness
          slab_thickness   nslab
          ________________________|_

                 nclad
          __________________________
          <------------------------>
                   w_sim

    """

    wavelength: float
    wg_width: float
    wg_thickness: float
    ncore: MaterialSpec
    nclad: MaterialSpec
    dn_dict: Optional[Dict] = None
    slab_thickness: float
    t_box: float = 2.0
    t_clad: float = 2.0
    xmargin: float = 1.0
    resolution: Union[int, Tuple[int, int]] = 100
    nmodes: int = 4
    bend_radius: Optional[float] = None
    cache: bool = True
    precision: Precision = "single"
    filter_pol: Optional[FilterPol] = None

    loss_model: Optional[bool] = False
    sidewall_sigma: Optional[float] = 10 * nm
    sidewall_k: Optional[float] = 0.1
    top_sigma: Optional[float] = 10 * nm
    top_k: Optional[float] = 0.1

    class Config:
        """Config for Waveguide."""

        extra = Extra.allow

    @property
    def cache_path(self) -> Optional[PathType]:
        return get_modes_path()

    @property
    def t_sim(self) -> float:
        return self.t_box + self.wg_thickness + self.t_clad

    @property
    def settings(self):
        return SETTINGS

    @property
    def w_sim(self) -> float:
        return self.wg_width + 2 * self.xmargin

    @property
    def filepath(self) -> Optional[pathlib.Path]:
        if not self.cache:
            return
        cache = pathlib.Path(self.cache_path)
        cache.mkdir(exist_ok=True, parents=True)

        settings = [
            f"{setting}={clean_value_name(getattr(self, setting))}"
            for setting in sorted(self.settings)
        ]

        named_args_string = "_".join(settings)
        h = hashlib.md5(named_args_string.encode()).hexdigest()[:8]
        return cache / f"{h}.npz"

    def get_ncore(self, wavelength: Optional[float] = None) -> float:
        wavelength = wavelength or self.wavelength
        return get_material_index(self.ncore, wavelength)

    def get_nclad(self, wavelength: Optional[float] = None) -> float:
        wavelength = wavelength or self.wavelength
        return get_material_index(self.nclad, wavelength)

    def get_n(self, Y, Z):
        """Return index matrix for a waveguide.

        Args:
            Y: 2D array.
            Z: 2D array.

        """
        w = self.wg_width
        ncore = self.get_ncore()
        nclad = self.get_nclad()
        t_box = self.t_box
        wg_thickness = self.wg_thickness
        slab_thickness = self.slab_thickness
        t_clad = self.t_clad

        inds_core = (
            (-w / 2 <= Y) & (Y <= w / 2) & (Z >= t_box) & (Z <= t_box + wg_thickness)
        )
        inds_slab = (Z >= t_box) & (Z <= t_box + slab_thickness)

        complex_solver = False
        mat_dtype = np.float32
        if isinstance(ncore, complex) or isinstance(nclad, complex):
            complex_solver = True
        elif self.dn_dict is not None:
            complex_solver = True
        elif self.loss_model:
            complex_solver = True
        if complex_solver:
            mat_dtype = np.complex128 if self.precision == "double" else np.complex64
        elif self.precision == "double":
            mat_dtype = np.float64

        n = np.ones_like(Y, dtype=mat_dtype) * nclad
        n[
            (-w / 2 - t_clad / 2 <= Y)
            & (Y <= w / 2 + t_clad / 2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness + t_clad)
        ] = nclad
        n[(Z <= 1.0 + slab_thickness + t_clad)] = nclad
        n[inds_core] = ncore
        n[inds_slab] = ncore if slab_thickness else nclad

        if self.loss_model:
            inds_top = (
                (Z >= t_box + wg_thickness - self.top_sigma / 2)
                & (Z <= t_box + wg_thickness + self.top_sigma / 2)
                & (-w / 2 <= Y)
                & (Y <= w / 2)
            )
            inds_top_slab_left = (
                (Z >= t_box + slab_thickness - self.top_sigma / 2)
                & (Z <= t_box + slab_thickness + self.top_sigma / 2)
                & (-w / 2 >= Y)
            )
            inds_top_slab_right = (
                (Z >= t_box + slab_thickness - self.top_sigma / 2)
                & (Z <= t_box + slab_thickness + self.top_sigma / 2)
                & (Y >= w / 2)
            )
            inds_sidewall_left = (
                (Z >= t_box + slab_thickness)
                & (Z <= t_box + wg_thickness)
                & (Y >= -w / 2 - self.sidewall_sigma / 2)
                & (Y <= -w / 2 + self.sidewall_sigma / 2)
            )
            inds_sidewall_right = (
                (Z >= t_box + slab_thickness)
                & (Z <= t_box + wg_thickness)
                & (Y >= w / 2 - self.sidewall_sigma / 2)
                & (Y <= w / 2 + self.sidewall_sigma / 2)
            )
            n[inds_top] += 1j * self.top_k
            n[inds_top_slab_left] += 1j * self.top_k
            n[inds_top_slab_right] += 1j * self.top_k
            n[inds_sidewall_left] += 1j * self.sidewall_k
            n[inds_sidewall_right] += 1j * self.sidewall_k

        if self.dn_dict is not None:
            dn = griddata(
                (self.dn_dict["x"], self.dn_dict["y"]),
                self.dn_dict["dn"],
                (Y, Z),
                method="cubic",
                fill_value=0.0,
            )
            n[inds_core] += dn[inds_core]
            n[inds_slab] += dn[inds_slab]

        return n

    def plot_index(self, func=None) -> None:
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution[0]
            if isinstance(self.resolution, tuple)
            else self.resolution,
            self.resolution[1]
            if isinstance(self.resolution, tuple)
            else self.resolution,
        )

        nx = self.get_n(
            Xx,
            Yx,
        )
        if func is None:
            plot(Xx, Yx, nx)
        else:
            plot(Xx, Yx, func(nx))
        plt.show()

    def compute_modes(
        self,
        overwrite: bool = False,
        with_fields: bool = True,
        isolate: bool = False,
    ) -> None:
        """Compute modes.

        Args:
            overwrite: overwrite file cache.
            with_fields: include field data.
            isolate: whether to run the solver in this interpreter (False) or a separate one (True)
            temp_dir: if isolate, which directory to save temporary files to
        """
        if hasattr(self, "neffs") and not overwrite:
            return

        wavelength = self.wavelength
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution[0]
            if isinstance(self.resolution, tuple)
            else self.resolution,
            self.resolution[1]
            if isinstance(self.resolution, tuple)
            else self.resolution,
        )

        nx = self.get_n(
            Xx,
            Yx,
        )
        ny = self.get_n(
            Xy,
            Yy,
        )
        nz = self.get_n(
            Xz,
            Yz,
        )
        self.nx, self.ny, self.nz = nx, ny, nz
        self.Xx, self.Yx, self.Xy, self.Yy, self.Xz, self.Yz = Xx, Yx, Xy, Yy, Xz, Yz

        if self.cache and self.filepath and self.filepath.exists():
            data = np.load(self.filepath)

            if with_fields:
                self.Ex = data["Ex"]
                self.Ey = data["Ey"]
                self.Ez = data["Ez"]
                self.Hx = data["Hx"]
                self.Hy = data["Hy"]
                self.Hz = data["Hz"]
            self.neffs = data["neffs"]
            logger.info(f"load {self.filepath} mode data from file cache.")
            return

        if isolate:
            # TODO: use process
            import pickle

            # Setup paths
            temp_dir = Path.cwd() / "temp"
            temp_dir.mkdir(exist_ok=True, parents=True)
            args_file_str = "args"
            argsfile = temp_dir / args_file_str
            argsfile = argsfile.with_suffix(".pkl")
            script_file_str = "script"
            scriptfile = temp_dir / script_file_str
            scriptfile = scriptfile.with_suffix(".py")
            outputs_file_str = "outputs"
            outputsfile = temp_dir / outputs_file_str
            outputsfile = outputsfile.with_suffix(".pkl")
            arguments_dict = {
                "nx": nx,
                "ny": ny,
                "nz": nz,
                "x": x,
                "y": y,
                "SPEED_OF_LIGHT": SPEED_OF_LIGHT,
                "wavelength": wavelength,
                "nmodes": self.nmodes,
                "bend_radius": self.bend_radius,
                "bend_axis": 1,
                "angle_theta": 0.0,
                "angle_phi": 0.0,
                "num_pml": (0, 0),
                "target_neff": self.get_ncore(wavelength),
                "sort_by": "largest_neff",
                "precision": self.precision,
                "filter_pol": self.filter_pol,
            }
            with open(argsfile, "wb") as outp:
                pickle.dump(arguments_dict, outp, pickle.HIGHEST_PROTOCOL)
            # Write execution file
            script_lines = [
                "import pickle\n",
                "import numpy as np\n",
                "from types import SimpleNamespace\n",
                "from tidy3d.plugins.mode.solver import compute_modes\n\n",
                'if __name__ == "__main__":\n\n',
                f"\twith open({str(argsfile)!r}, 'rb') as inp:\n",
                "\t\targuments_dict = pickle.load(inp)\n\n",
            ]
            script_lines.extend(
                f"\t{key} = arguments_dict[{key!r}]\n" for key in arguments_dict
            )
            script_lines.extend(
                [
                    "\t((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (\n",
                    "\t    x.squeeze()\n",
                    "\t    for x in compute_modes(\n",
                    "\t        eps_cross=[nx**2, ny**2, nz**2],\n",
                    "\t        coords=[x, y],\n",
                    "\t        freq=SPEED_OF_LIGHT / (wavelength * 1e-6),\n",
                    "\t        mode_spec=SimpleNamespace(\n",
                    "\t            num_modes=nmodes,\n",
                    "\t            bend_radius=bend_radius,\n",
                    "\t            bend_axis=bend_axis,\n",
                    "\t            angle_theta=angle_theta,\n",
                    "\t            angle_phi=angle_phi,\n",
                    "\t            num_pml=num_pml,\n",
                    "\t            target_neff=target_neff,\n",
                    "\t            sort_by=sort_by,\n",
                    "\t            precision=precision,\n",
                    "\t            filter_pol=filter_pol,\n",
                    "\t        ),\n",
                    "\t    )\n",
                    "\t)\n",
                ]
            )
            script_lines.extend(
                [
                    f"\toutputsfile = {str(outputsfile)!r}\n",
                    "\toutputs_dict = {\n",
                    '\t\t    "Ex": Ex,\n',
                    '\t\t    "Ey": Ey,\n',
                    '\t\t    "Ez": Ez,\n',
                    '\t\t    "Hx": Hx,\n',
                    '\t\t    "Hy": Hy,\n',
                    '\t\t    "Hz": Hz,\n',
                    '\t\t    "neffs": neffs,\n',
                    "\t\t}\n",
                    '\twith open(outputsfile, "wb") as outp:\n',
                    "\t\t    pickle.dump(outputs_dict, outp, pickle.HIGHEST_PROTOCOL)\n",
                ]
            )
            with open(scriptfile, "w") as script_file_obj:
                script_file_obj.writelines(script_lines)
            with subprocess.Popen(
                ["python", scriptfile],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as proc:
                if not proc.stderr:
                    while not outputsfile.exists():
                        print(proc.stdout.read().decode())
                        print(proc.stderr.read().decode())
                        sys.stdout.flush()
                        sys.stderr.flush()
                        time.sleep(1)
                logger.info(f"python {scriptfile}")

            with open(outputsfile, "rb") as inp:
                outputs_dict = pickle.load(inp)

            Ex = outputs_dict["Ex"]
            Ey = outputs_dict["Ey"]
            Ez = outputs_dict["Ez"]
            Hx = outputs_dict["Hx"]
            Hy = outputs_dict["Hy"]
            Hz = outputs_dict["Hz"]
            neffs = outputs_dict["neffs"]

            import shutil

            shutil.rmtree(temp_dir)

        else:  # legacy
            ((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (
                x.squeeze()
                for x in compute_modes(
                    eps_cross=[nx**2, ny**2, nz**2],
                    coords=[x, y],
                    freq=SPEED_OF_LIGHT / (wavelength * 1e-6),
                    mode_spec=SimpleNamespace(
                        num_modes=self.nmodes,
                        bend_radius=self.bend_radius,
                        bend_axis=1,
                        angle_theta=0.0,
                        angle_phi=0.0,
                        num_pml=(0, 0),
                        target_neff=self.get_ncore(wavelength),
                        sort_by="largest_neff",
                        precision=self.precision,
                        filter_pol=self.filter_pol,
                    ),
                )
            )

        self.Ex, self.Ey, self.Ez = Ex, Ey, Ez
        self.Hx, self.Hy, self.Hz = Hx, Hy, Hz
        self.neffs = neffs

        if with_fields:
            data = dict(
                Ex=self.Ex,
                Ey=self.Ey,
                Ez=self.Ez,
                Hx=self.Hx,
                Hy=self.Hy,
                Hz=self.Hz,
                neffs=self.neffs,
            )
        else:
            data = dict(neffs=self.neffs)
        if self.filepath:
            np.savez_compressed(self.filepath, **data)
            logger.info(f"write {self.filepath} mode data to file cache.")

    def compute_mode_properties(self) -> Tuple[List[float], List[float], List[float]]:
        """Computes mode areas, fraction_te and fraction_tm."""
        if not hasattr(self, "neffs"):
            self.compute_modes()
        mode_areas = []
        fraction_te = []
        fraction_tm = []

        for mode_index in range(self.nmodes):
            e_fields = (
                self.Ex[..., mode_index],
                self.Ey[..., mode_index],
                self.Ez[..., mode_index],
            )
            h_fields = (
                self.Hx[..., mode_index],
                self.Hy[..., mode_index],
                self.Hz[..., mode_index],
            )

            areas_e = [np.sum(np.abs(e) ** 2) for e in e_fields]
            areas_e /= np.sum(areas_e)
            areas_e *= 100

            areas_h = [np.sum(np.abs(h) ** 2) for h in h_fields]
            areas_h /= np.sum(areas_h)
            areas_h *= 100

            fraction_te.append(areas_e[0] / (areas_e[0] + areas_e[1]))
            fraction_tm.append(areas_e[1] / (areas_e[0] + areas_e[1]))

            areas = areas_e.tolist()
            areas.extend(areas_h)
            mode_areas.append(areas)

        self.mode_areas = mode_areas
        self.fraction_te = fraction_te
        self.fraction_tm = fraction_tm
        return mode_areas, fraction_te, fraction_tm

    def plot_Ex(self, mode_index: int = 0, plot_power: bool = True, **kwargs) -> None:
        """Plot mode.

        Args:
            mode_index: 0 is the fundamental mode.
            plot_power: if False plots field.

        Keyword Args:
            num_levels: for plot.
            n_cmap: refractive index color map.
            mode_cmap: mode color map.
            axes: "xy".
            title: for the plot.
            normalize_mode: divide by maximum value.
        """
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ex = self.nx, self.neffs, self.Ex
        neff_, Ex_ = np.real(neffs[mode_index]), Ex[..., mode_index]
        mode = np.abs(Ex_) ** 2 if plot_power else Ex_.real

        plot(
            self.Xx,
            self.Yx,
            nx,
            mode=mode,
            title=f"Ex::{neff_:.3f}",
            plot_power=plot_power,
            **kwargs,
        )
        plt.show()

    def plot_Ey(self, mode_index: int = 0) -> None:
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ey = self.nx, self.neffs, self.Ey
        neff_, Ey_ = np.real(neffs[mode_index]), Ey[..., mode_index]
        plot(self.Xx, self.Yx, nx, mode=np.abs(Ey_) ** 2, title=f"Ey::{neff_:.3f}")
        plt.show()

    def _ipython_display_(self) -> None:
        """Show index in matplotlib for Jupyter Notebooks."""
        self.plot_index()

    def __repr__(self) -> str:
        """Show waveguide name."""
        return ", \n".join([f"{k} = {getattr(self, k)!r}" for k in self.settings])

    def get_overlap(
        self, wg: Waveguide, mode_index1: int = 0, mode_index2: int = 0
    ) -> float:
        """Returns mode overlap integral.

        Args:
            wg: other waveguide.
        """
        wg1 = self
        wg2 = wg

        wg1.compute_modes()
        wg2.compute_modes()

        return np.sum(
            np.conj(wg1.Ex[..., mode_index1]) * wg2.Hy[..., mode_index2]
            - np.conj(wg1.Ey[..., mode_index1]) * wg2.Hx[..., mode_index2]
            + wg2.Ex[..., mode_index2] * np.conj(wg1.Hy[..., mode_index1])
            - wg2.Ey[..., mode_index2] * np.conj(wg1.Hx[..., mode_index1])
        )

    def get_loss(self):
        """Returns loss for computed modes in dB/cm."""
        if not hasattr(self, "neffs"):
            self.compute_modes()
        wavelength = self.wavelength * 1e-6  # convert to m
        alphas = 4 * np.pi * np.imag(self.neffs) / wavelength  # lin/m loss
        return 10 * np.log10(np.exp(1)) * alphas * 1e-2  # dB/cm loss


class WaveguideCoupler(Waveguide):
    """Waveguide coupler Model.

    Parameters:
        wavelength: (um).
        wg_width1: left waveguide width in um.
        wg_width2: right waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um. Can be a single number or tuple (x, y).
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
        cache: True uses file cache from PDK.modes_path. False skips cache.

    ::

            wg_width1     wg_width2
            <------->     <------->
             _______   |   _______   __
            |       |  |  |       | |
            |       |  |  |       | |
            |       |_____|       | | wg_thickness
            |slab_thickness       | |
            |_____________________| |__
                    <----->
                      gap
            <--------------------->
                     w_sim
    """

    wg_width: Optional[float] = None
    wg_width1: float
    wg_width2: float
    gap: float

    @property
    def w_sim(self):
        return self.wg_width1 + self.wg_width2 + self.gap + 2 * self.xmargin

    def get_n(self, Y, Z):
        """Return index matrix for a waveguide coupler.

        Args:
            Y: 2D array.
            Z: 2D array.
        """
        w1 = self.wg_width1
        w2 = self.wg_width2
        gap = self.gap
        ncore = self.get_ncore()
        nclad = self.get_nclad()
        t_box = self.t_box
        wg_thickness = self.wg_thickness
        slab_thickness = self.slab_thickness
        t_clad = self.t_clad

        n = np.ones_like(Y) * nclad
        n[(Z <= 1.0 + slab_thickness + t_clad)] = nclad
        n[
            (-w1 - gap / 2 <= Y)
            & (Y <= -gap / 2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness)
        ] = ncore
        n[
            (gap / 2 <= Y)
            & (Y <= gap / 2 + w2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness)
        ] = ncore
        n[(Z >= t_box) & (Z <= t_box + slab_thickness)] = (
            ncore if slab_thickness else nclad
        )
        return n

    @property
    def settings(self):
        return SETTINGS_COUPLER

    def find_coupling(self, power_ratio: float = 1.0) -> float:
        """Returns the coupling length (um) of the directional coupler to achieve power_ratio, where 1 means 100% power transfer."""
        if not hasattr(self, "neffs"):
            self.compute_modes()
        neff1 = self.neffs[0]
        neff2 = self.neffs[1]
        dneff = (neff1 - neff2).real
        return self.wavelength / (np.pi * dneff) * np.arcsin(np.sqrt(power_ratio))


@pydantic.validate_arguments
def find_coupling_vs_gap(
    gap1: float = 0.2,
    gap2: float = 0.4,
    steps: int = 12,
    nmodes: int = 4,
    wavelength: float = 1.55,
    **kwargs,
) -> pd.DataFrame:
    """Returns coupling vs gap pandas DataFrame.

    Args:
        gap1: starting gap in um.
        gap2: end gap in um.
        steps: number of steps.
        nmodes: number of modes.
        wavelength: wavelength (um).

    Keyword Args:
        wg_width: waveguide width.
        wg_width1: optional left waveguide width in um.
        wg_width2: optional right waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um. Can be a single number or tuple (x, y).
        bend_radius: optional bend radius (um).
        cache: True uses file cache from PDK.modes_path. False skips cache.

    """

    power_ratio = 1
    gaps = np.linspace(gap1, gap2, steps)
    ne = []
    no = []
    dn = []
    lc = []

    for gap in tqdm(gaps):
        wg = WaveguideCoupler(
            gap=gap,
            nmodes=nmodes,
            wavelength=wavelength,
            **kwargs,
        )

        wg.compute_modes()
        neff1 = wg.neffs[0].real
        neff2 = wg.neffs[1].real
        dneff = (neff1 - neff2).real
        coupling = wg.wavelength / (np.pi * dneff) * np.arcsin(np.sqrt(power_ratio))

        ne.append(neff1)
        no.append(neff2)
        dn.append(neff1 - neff2)
        lc.append(coupling)

    return pd.DataFrame(dict(gap=gaps, ne=ne, no=no, lc=lc, dn=dn))


def sweep_bend_loss(
    bend_radius_min: float = 2.0,
    bend_radius_max: float = 5,
    steps: int = 4,
    mode_index: int = 0,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """Returns overlap integral squared for the bend mode mismatch loss.

    The loss is squared because you hit the bend loss twice
    (from bend to straight and from straight to bend).

    Args:
        bend_radius_min: min bend radius (um).
        bend_radius_max: max bend radius (um).
        steps: number of steps.
        mode_index: where 0 is the fundamental mode.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
    """
    r = np.linspace(bend_radius_min, bend_radius_max, steps)
    integral = np.zeros_like(r)

    wg = Waveguide(**kwargs)

    for i, radius in tqdm(enumerate(r)):
        wg_bent = Waveguide(bend_radius=radius, **kwargs)
        wg.get_overlap(wg_bent, mode_index1=mode_index, mode_index2=mode_index)

        # normalized overlap integral
        integral[i] = np.abs(
            wg.get_overlap(wg_bent, mode_index, mode_index) ** 2
            / wg.get_overlap(wg, mode_index, mode_index)
            / wg.get_overlap(wg_bent, mode_index, mode_index)
        )

    return r, integral**2


def sweep_neff(
    wavelength: float = 1.55,
    thicknesses: Tuple[float, ...] = (220 * nm,),
    widths: Tuple[float, ...] = (500 * nm,),
    mode_index: int = 0,
    **kwargs,
) -> pd.DataFrame:
    """Sweep waveguide width and compute effective index.

    Args:
        wavelength: (um).
        thicknesses: in um.
        widths: in um.
        mode_index: integer, where 0 is the fundamental mode.

    Keyword Args:
        mode_index: integer.
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """
    widths_thicknesses = list(it.product(widths, thicknesses))

    neff = np.zeros(len(widths_thicknesses))
    w = np.zeros(len(widths_thicknesses))
    t = np.zeros(len(widths_thicknesses))

    for i, (wg_width, wg_thickness) in enumerate(tqdm(widths_thicknesses)):
        wg = Waveguide(
            wg_width=wg_width,
            wg_thickness=wg_thickness,
            wavelength=wavelength,
            **kwargs,
        )
        wg.compute_modes()
        wg.compute_mode_properties()
        neff[i] = np.real(wg.neffs[mode_index])
        w[i] = wg_width
        t[i] = wg_thickness

    return pd.DataFrame(dict(neff=neff, widths=w, thickness=t))


def group_index(
    wavelength: float, wavelength_step: float = 0.01, mode_index: int = 0, **kwargs
) -> float:
    """Returns group_index.

    Args:
        wavelength: (um).
        wavelength_step: in um.
        mode_index: integer, where 0 is the fundamental mode.

    Keyword Args:
        wg_width: waveguide width.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """
    wc = Waveguide(wavelength=wavelength, **kwargs)
    wf = Waveguide(
        wavelength=wavelength + wavelength_step,
        **kwargs,
    )
    wb = Waveguide(
        wavelength=wavelength - wavelength_step,
        **kwargs,
    )
    wc.compute_modes()
    wb.compute_modes()
    wf.compute_modes()

    nc = np.real(wc.neffs[mode_index])
    nb = np.real(wb.neffs[mode_index])
    nf = np.real(wf.neffs[mode_index])
    return nc - wavelength * (nf - nb) / (2 * wavelength_step)


def sweep_group_index(
    wavelength: float = 1.55,
    thicknesses: Tuple[float, ...] = (220 * nm,),
    widths: Tuple[float, ...] = (500 * nm,),
    **kwargs,
) -> pd.DataFrame:
    """Sweep waveguide width and compute group index.

    Args:
        wavelength: (um).
        thicknesses: in um.
        widths: in um.

    Keyword Args:
        mode_index: integer.
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """
    widths_thicknesses = list(it.product(widths, thicknesses))

    ng = np.zeros(len(widths_thicknesses))
    w = np.zeros(len(widths_thicknesses))
    t = np.zeros(len(widths_thicknesses))

    for i, (wg_width, wg_thickness) in enumerate(tqdm(widths_thicknesses)):
        ng[i] = group_index(
            wavelength=wavelength,
            wg_width=wg_width,
            wg_thickness=wg_thickness,
            **kwargs,
        )
        w[i] = wg_width
        t[i] = wg_thickness

    return pd.DataFrame(dict(ng=ng, widths=w, thickness=t))


def sweep_width(
    width1: float = 200 * nm,
    width2: float = 1000 * nm,
    steps: int = 12,
    nmodes: int = 4,
    **kwargs,
) -> pd.DataFrame:
    """Sweep waveguide width and compute effective index.

    Returns pandas dataframe with effective index (neff) and fraction_te.

    Args:
        width1: starting waveguide width in um.
        width2: end waveguide width in um.
        steps: number of points.
        nmodes: number of modes to compute.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """
    width = np.linspace(width1, width2, steps)
    neff = {}
    for mode_number in range(nmodes):
        neff[f"neff{mode_number}"] = []
        neff[f"fraction_te{mode_number}"] = []

    for wg_width in tqdm(width):
        wg = Waveguide(nmodes=nmodes, wg_width=wg_width, **kwargs)
        wg.compute_modes()
        wg.compute_mode_properties()

        for mode_number in range(nmodes):
            n = np.real(wg.neffs[mode_number])
            fraction_te = wg.fraction_te[mode_number]
            neff[f"neff{mode_number}"].append(n)
            neff[f"fraction_te{mode_number}"].append(fraction_te)

    df = pd.DataFrame(neff)
    df["width"] = width
    return df


def plot_sweep_width(
    width1: float = 200 * nm,
    width2: float = 1000 * nm,
    steps: int = 12,
    nmodes: int = 4,
    cmap: str = "magma",
    **kwargs,
) -> pd.DataFrame:
    """Plots and Returns Dataframe with waveguide width sweep results.

    Returns pandas dataframe with effective index (neff) and fraction_te.

    Args:
        width1: starting waveguide width in um.
        width2: end waveguide width in um.
        steps: number of points.
        nmodes: number of modes to compute.
        cmap: colormap for the TE fraction.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um. Can be a single number or tuple (x, y).
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """
    width = np.linspace(width1, width2, steps)
    neff = {}
    for mode_number in range(nmodes):
        neff[f"neff{mode_number}"] = []
        neff[f"fraction_te{mode_number}"] = []

    for wg_width in tqdm(width):
        wg = Waveguide(nmodes=nmodes, wg_width=wg_width, **kwargs)
        wg.compute_modes()
        wg.compute_mode_properties()
        for mode_number in range(nmodes):
            n = np.real(wg.neffs[mode_number])
            fraction_te = wg.fraction_te[mode_number]
            plt.scatter(wg_width, n, c=fraction_te, vmin=0, vmax=1, cmap=cmap)

            neff[f"neff{mode_number}"].append(n)
            neff[f"fraction_te{mode_number}"].append(fraction_te)

    for mode_number in range(nmodes):
        plt.plot(width, neff[f"neff{mode_number}"], c="gray")

    plt.colorbar().set_label("TE-Fraction")
    plt.xlabel("width (um)")
    plt.ylabel("neff")
    df = pd.DataFrame(neff)
    df["width"] = width
    return df


__all__ = (
    "Waveguide",
    "plot_sweep_width",
    "si",
    "sin",
    "sio2",
    "sweep_bend_loss",
    "sweep_width",
    "sweep_neff",
    "sweep_group_index",
    "group_index",
)


if __name__ == "__main__":
    # wg = Waveguide(
    #     wavelength=1.55,
    #     wg_width=0.5,
    #     wg_thickness=0.22,
    #     slab_thickness=0.0,
    #     ncore="si",
    #     nclad="sio2",
    # )
    wg = WaveguideCoupler(
        wavelength=1550 * nm,
        gap=200 * nm,
        wg_width1=500 * nm,
        wg_width2=500 * nm,
        wg_thickness=220 * nm,
        slab_thickness=0 * nm,
        ncore="si",
        nclad="sio2",
    )
    print(wg.filepath)

    # wg = Waveguide(
    #     nmodes=2,
    #     wg_width=500 * nm,
    #     wavelength=1.55,
    #     wg_thickness=220 * nm,
    #     slab_thickness=90 * nm,
    #     ncore="si",
    #     nclad="sio2",
    #     loss_model=True,
    #     sidewall_k=1e-4,
    #     top_k=1e-4,
    #     sidewall_sigma=20 * nm,
    #     top_sigma=20 * nm,
    #     resolution=400,
    #     cache=True,
    #     precision="double",
    # )
    # wg.plot_index()
    # wg.plot_index(func=np.imag)
    # wg.compute_mode_properties()

    # wg.compute_modes()
    # print(wg.neffs)
    # print(wg.get_loss())
    # wg.plot_Ex()
    # wg.compute_mode_properties()
    # for mode_number in range(nmodes):
    #     n = np.real(wg.neffs[mode_number])
    #     fraction_te = wg.fraction_te[mode_number]
    #     plt.scatter(wg_width, n, c=fraction_te, vmin=0, vmax=1, cmap=cmap)
    # n[inds_top_slab_left] += self.top_k
    # n[inds_top_slab_right] += self.top_k
    # n[inds_sidewall_left] += self.sidewall_k
    # n[inds_sidewall_right] += self.sidewall_k



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from scipy.interpolate import NearestNDInterpolator
from shapely.geometry import LineString, MultiPolygon, Point, Polygon
from shapely.ops import unary_union

import gdsfactory as gf
from gdsfactory.simulation.gmsh.mesh import mesh_from_polygons
from gdsfactory.simulation.gmsh.parse_component import (
    create_2D_surface_interface,
    merge_by_material_func,
    process_buffers,
)
from gdsfactory.simulation.gmsh.parse_gds import cleanup_component, to_polygons
from gdsfactory.simulation.gmsh.parse_layerstack import (
    list_unique_layerstack_z,
    order_layerstack,
)
from gdsfactory.technology import LayerStack
from gdsfactory.typings import ComponentOrReference


def get_u_bounds_polygons(
    polygons: Union[MultiPolygon, List[Polygon]],
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
):
    """Performs the bound extraction given a (Multi)Polygon or [Polygon] and cross-sectional line coordinates.

    Args:
        layer_polygons_dict: dict containing layernames: shapely polygons pairs
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end.

    Returns: list of bounding box coordinates (u1,u2)) in xsection line coordinates (distance from xsection_bounds[0]).
    """
    line = LineString(xsection_bounds)
    linestart = Point(xsection_bounds[0])

    return_list = []
    for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
        initial_settings = np.seterr()
        np.seterr(invalid="ignore")
        intersection = polygon.intersection(line)
        np.seterr(**initial_settings)
        if intersection:
            for entry in (
                intersection.geoms if hasattr(intersection, "geoms") else [intersection]
            ):
                bounds = entry.bounds
                p1 = Point([bounds[0], bounds[1]])
                p2 = Point([bounds[2], bounds[3]])
                return_list.append([linestart.distance(p1), linestart.distance(p2)])
    return return_list


def get_u_bounds_layers(
    layer_polygons_dict: Dict[Tuple(str, str, str), MultiPolygon],
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
):
    """Given a layer_polygons_dict and two coordinates (x1,y1), (x2,y2), computes the \
        bounding box(es) of each layer in the xsection coordinate system (u).

    Args:
        layer_polygons_dict: dict containing layernames: shapely polygons pairs
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end.

    Returns: Dict containing layer(list pairs, with list a list of bounding box coordinates (u1,u2))
        in xsection line coordinates.
    """
    bounds_dict = {}
    for layername, (
        gds_layername,
        next_layername,
        polygons,
        next_polygons,
    ) in layer_polygons_dict.items():
        bounds_dict[layername] = []
        bounds = get_u_bounds_polygons(polygons, xsection_bounds)
        next_bounds = get_u_bounds_polygons(next_polygons, xsection_bounds)
        if bounds:
            bounds_dict[layername] = (
                gds_layername,
                next_layername,
                bounds,
                next_bounds,
            )

    return bounds_dict


def get_uz_bounds_layers(
    layer_polygons_dict: Dict[str, Tuple[str, MultiPolygon, MultiPolygon]],
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    layerstack: LayerStack,
    u_offset: float = 0.0,
):
    """Given a component and layer stack, computes the bounding box(es) of each \
            layer in the xsection coordinate system (u,z).

    Args:
        component: Component or ComponentReference.
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end

    Returns: Dict containing layer: polygon pairs, with (u1,u2) in xsection line coordinates
    """
    # Get in-plane cross-sections
    inplane_bounds_dict = get_u_bounds_layers(layer_polygons_dict, xsection_bounds)

    outplane_bounds_dict = {}

    layer_dict = layerstack.to_dict()

    # Remove empty entries
    inplane_bounds_dict = {
        key: value for (key, value) in inplane_bounds_dict.items() if value
    }

    for layername, (
        gds_layername,
        next_layername,
        inplane_bounds_list,
        next_inplane_bounds_list,
    ) in inplane_bounds_dict.items():
        outplane_polygons_list = []
        for inplane_bounds, next_inplane_bounds in zip(
            inplane_bounds_list, next_inplane_bounds_list
        ):
            zmin = layer_dict[layername]["zmin"]
            zmax = layer_dict[next_layername]["zmin"]

            # Get bounding box
            umin_zmin = np.min(inplane_bounds) + u_offset
            umax_zmin = np.max(inplane_bounds) + u_offset
            umin_zmax = np.min(next_inplane_bounds) + u_offset
            umax_zmax = np.max(next_inplane_bounds) + u_offset

            points = [
                [umin_zmin, zmin],
                [umin_zmax, zmax],
                [umax_zmax, zmax],
                [umax_zmin, zmin],
            ]
            outplane_polygons_list.append(Polygon(points))

        outplane_bounds_dict[layername] = (gds_layername, outplane_polygons_list)

    return outplane_bounds_dict


def uz_xsection_mesh(
    component: ComponentOrReference,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    layerstack: LayerStack,
    resolutions: Optional[Dict] = None,
    mesh_scaling_factor: float = 1.0,
    default_resolution_min: float = 0.01,
    default_resolution_max: float = 0.5,
    background_tag: Optional[str] = None,
    background_padding: Tuple[float, float, float, float] = (2.0, 2.0, 2.0, 2.0),
    filename: Optional[str] = None,
    global_meshsize_array: Optional[np.array] = None,
    global_meshsize_interpolant_func: Optional[callable] = NearestNDInterpolator,
    extra_shapes_dict: Optional[OrderedDict] = None,
    merge_by_material: Optional[bool] = False,
    interface_surfaces: Optional[Dict[str, Tuple(float, float)]] = None,
    round_tol: int = 3,
    simplify_tol: float = 1e-2,
    u_offset: float = 0.0,
    **kwargs,
):
    """Mesh uz cross-section of component along line u = [[x1,y1] , [x2,y2]].

    Args:
        component (Component): gdsfactory component to mesh
        xsection_bounds (Tuple): Tuple [[x1,y1] , [x2,y2]] parametrizing the line u
        layerstack (LayerStack): gdsfactory LayerStack to parse
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}} to roughly control mesh refinement
        mesh_scaling_factor (float): factor multiply mesh geometry by
        default_resolution_min (float): gmsh minimal edge length
        default_resolution_max (float): gmsh maximal edge length
        background_tag (str): name of the background layer to add (default: no background added)
        background_padding (Tuple): [xleft, ydown, xright, yup] distances to add to the components and to fill with background_tag
        filename (str, path): where to save the .msh file
        global_meshsize_array: np array [x,y,z,lc] to parametrize the mesh
        global_meshsize_interpolant_func: interpolating function for global_meshsize_array
        extra_shapes_dict: Optional[OrderedDict] = OrderedDict of {key: geo} with key a label and geo a shapely (Multi)Polygon or (Multi)LineString of extra shapes to override component
        merge_by_material: boolean, if True will merge polygons from layers with the same layer.material. Physical keys will be material in this case.
        round_tol: during gds --> mesh conversion cleanup, number of decimal points at which to round the gdsfactory/shapely points before introducing to gmsh
        simplify_tol: during gds --> mesh conversion cleanup, shapely "simplify" tolerance (make it so all points are at least separated by this amount)
        u_offset: quantity to add to the "u" coordinates, useful to convert back to x or y if parallel to those axes
    """
    interface_surfaces = interface_surfaces or {}

    # Fuse and cleanup polygons of same layer in case user overlapped them
    layer_polygons_dict = cleanup_component(
        component, layerstack, round_tol, simplify_tol
    )

    # GDS polygons to simulation polygons
    buffered_layer_polygons_dict, buffered_layerstack = process_buffers(
        layer_polygons_dict, layerstack
    )

    # simulation polygons to u-z coordinates along cross-sectional line
    bounds_dict = get_uz_bounds_layers(
        buffered_layer_polygons_dict, xsection_bounds, buffered_layerstack, u_offset
    )

    # u-z coordinates to gmsh-friendly polygons
    # Remove terminal layers and merge polygons
    layer_order = order_layerstack(layerstack)  # gds layers
    shapes = OrderedDict() if extra_shapes_dict is None else extra_shapes_dict
    for layername in layer_order:
        current_shapes = []
        for _, (gds_name, bounds) in bounds_dict.items():
            if gds_name == layername:
                layer_shapes = list(bounds)
                current_shapes.append(MultiPolygon(to_polygons(layer_shapes)))
        shapes[layername] = unary_union(MultiPolygon(to_polygons(current_shapes)))

    # Add background polygon
    if background_tag is not None:
        # shapes[background_tag] = bounds.buffer(background_padding[0])
        # bounds = unary_union(list(shapes.values())).bounds
        zs = list_unique_layerstack_z(buffered_layerstack)
        zmin = np.min(zs)
        zmax = np.max(zs)
        shapes[background_tag] = Polygon(
            [
                [-1 * background_padding[0] + u_offset, zmin - background_padding[1]],
                [-1 * background_padding[0] + u_offset, zmax + background_padding[3]],
                [
                    np.linalg.norm(
                        np.array(xsection_bounds[1]) - np.array(xsection_bounds[0])
                    )
                    + background_padding[2]
                    + u_offset,
                    zmax + background_padding[3],
                ],
                [
                    np.linalg.norm(
                        np.array(xsection_bounds[1]) - np.array(xsection_bounds[0])
                    )
                    + background_padding[2]
                    + u_offset,
                    zmin - background_padding[1],
                ],
            ]
        )

    # Merge by material
    if merge_by_material:
        shapes = merge_by_material_func(shapes, layerstack)

    # Create interface surfaces
    reordered_shapes = OrderedDict()
    for label in shapes.keys():
        if interface_surfaces and label in interface_surfaces.keys():
            buffer_in, buffer_out, simplification = interface_surfaces[label]
            reordered_shapes[f"{label}Int"] = create_2D_surface_interface(
                shapes[label], buffer_in, buffer_out, simplification
            )
        reordered_shapes[label] = shapes[label]

    # Mesh
    return mesh_from_polygons(
        reordered_shapes,
        resolutions=resolutions,
        mesh_scaling_factor=mesh_scaling_factor,
        filename=filename,
        default_resolution_min=default_resolution_min,
        default_resolution_max=default_resolution_max,
        global_meshsize_array=global_meshsize_array,
        global_meshsize_interpolant_func=global_meshsize_interpolant_func,
    )


if __name__ == "__main__":
    from gdsfactory.pdk import get_layer_stack

    c = gf.component.Component()

    waveguide = c << gf.get_component(gf.components.straight_pin(length=10, taper=None))
    undercut = c << gf.get_component(
        gf.components.rectangle(
            size=(5.0, 5.0),
            layer="UNDERCUT",
            centered=True,
        )
    ).move(destination=[4, 0])
    c.show()

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "via_contact",
                # "undercut",
                "box",
                # "substrate",
                "clad",
                "metal1",
            )  # "slab90", "via_contact")#"via_contact") # "slab90", "core"
        }
    )

    resolutions = {}
    resolutions["si"] = {"resolution": 0.02, "distance": 0.1}
    resolutions["siInt"] = {"resolution": 0.0003, "distance": 0.1}
    # resolutions["sio2"] = {"resolution": 0.5, "distance": 1}
    resolutions["Aluminum"] = {"resolution": 0.1, "distance": 1}

    # resolutions = {}
    # resolutions["core"] = {"resolution": 0.02, "distance": 0.1}
    # resolutions["coreInt"] = {"resolution": 0.0001, "distance": 0.1}
    # resolutions["slab90"] = {"resolution": 0.05, "distance": 0.1}
    # # resolutions["sio2"] = {"resolution": 0.5, "distance": 1}
    # resolutions["via_contact"] = {"resolution": 0.1, "distance": 1}

    geometry = uz_xsection_mesh(
        c,
        [(4, -15), (4, 15)],
        filtered_layerstack,
        resolutions=resolutions,
        background_tag="Oxide",
        background_padding=(0, 0, 0, 0),
        filename="mesh.msh",
        round_tol=3,
        simplify_tol=1e-3,
        u_offset=-15,
    )

    import meshio

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=False):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points[:, :2] if prune_z else mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    meshio.write("mesh.xdmf", triangle_mesh)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.gdsdiff.gdsdiff import xor_polygons


def test_differences() -> None:
    straight = gf.partial(
        gf.components.straight,
        with_bbox=True,
        cladding_layers=None,
        add_pins=None,
        add_bbox=None,
    )
    c1 = straight(length=2)
    c2 = straight(length=3)
    c = xor_polygons(c1, c2, hash_geometry=False)
    area = c.area()
    assert area == 0.5, area


def test_no_differences() -> None:
    straight = gf.partial(
        gf.components.straight,
        with_bbox=True,
        cladding_layers=None,
        add_pins=None,
        add_bbox=None,
    )
    c1 = straight(length=2)
    c2 = straight(length=2)
    c = xor_polygons(c1, c2, hash_geometry=False)
    area = c.area()
    assert area == 0, area


if __name__ == "__main__":
    # test_no_differences()
    # test_differences()
    c1 = gf.components.mzi(length_x=3)
    c2 = gf.components.mzi(length_x=2)
    c = xor_polygons(c1, c2, hash_geometry=False)
    c.show(show_ports=True)



from __future__ import annotations

from typing import Dict, List, Union

from numpy import float64

from gdsfactory.port import Port


def flip(port: Port) -> Port:
    """Returns port copy with Flip Port orientation."""
    return port.flip()


def direction_ports_from_list_ports(optical_ports: List[Port]) -> Dict[str, List[Port]]:
    """Returns a dict of WENS ports."""
    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}
    for p in optical_ports:
        p.orientation = (p.orientation + 360.0) % 360
        if p.orientation <= 45.0 or p.orientation >= 315:
            direction_ports["E"].append(p)
        elif p.orientation <= 135.0 and p.orientation >= 45.0:
            direction_ports["N"].append(p)
        elif p.orientation <= 225.0 and p.orientation >= 135.0:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    for direction, list_ports in list(direction_ports.items()):
        if direction in ["E", "W"]:
            list_ports.sort(key=lambda p: p.y)

        if direction in ["S", "N"]:
            list_ports.sort(key=lambda p: p.x)

    return direction_ports


def check_ports_have_equal_spacing(list_ports: List[Port]) -> float64:
    """Returns port separation.

    Raises error if not constant.

    """
    if not isinstance(list_ports, list):
        raise ValueError(f"list_ports should be a list of ports, got {list_ports}")
    if not list_ports:
        raise ValueError("list_ports should not be empty")

    orientation = get_list_ports_angle(list_ports)
    if orientation in [0, 180]:
        xys = [p.y for p in list_ports]
    else:
        xys = [p.x for p in list_ports]

    seps = [round(abs(c2 - c1), 5) for c1, c2 in zip(xys[1:], xys[:-1])]
    different_seps = set(seps)
    if len(different_seps) > 1:
        raise ValueError("Ports should have the same separation. Got {different_seps}")

    return different_seps.pop()


def get_list_ports_angle(list_ports: List[Port]) -> Union[float64, int]:
    """Returns the orientation/angle (in degrees) of a list of ports."""
    if not list_ports:
        return None
    if len({p.orientation for p in list_ports}) > 1:
        raise ValueError(f"All port angles should be the same. Got {list_ports}")
    return list_ports[0].orientation


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.mmi1x2()
    d = direction_ports_from_list_ports(c.get_ports_list())
    c.show(show_ports=True)



"""read Sparameters from CSV file and returns sax model."""
from __future__ import annotations

import pathlib
from typing import Union

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from sax.typing_ import Float, Model
from typing_extensions import Literal

import gdsfactory as gf
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_lumerical,
    get_sparameters_path_meep,
    get_sparameters_path_tidy3d,
)

wl_cband = np.linspace(1.500, 1.600, 128)

PathType = Union[str, pathlib.Path]

Simulator = Literal["lumerical", "meep", "tidy3d"]


def model_from_npz(
    filepath: Union[PathType, np.ndarray],
    xkey: str = "wavelengths",
    xunits: float = 1,
    prefix: str = "s",
) -> Model:
    """Returns a SAX Sparameters Model from a npz file.

    The SAX Model is a function that returns a SAX SDict interpolated over wavelength.

    Args:
        filepath: CSV Sparameters path or pandas DataFrame.
        wl: wavelength to interpolate (um).
        xkey: key for wavelengths in file.
        xunits: x units in um from the loaded file (um). 1 means 1um.
        prefix: for the sparameters column names in file.

    """
    sp = np.load(filepath) if isinstance(filepath, (pathlib.Path, str)) else filepath
    keys = list(sp.keys())

    if xkey not in keys:
        raise ValueError(f"{xkey!r} not in {keys}")

    x = jnp.asarray(sp[xkey] * xunits)
    wl = jnp.asarray(wl_cband)

    # make sure x is sorted from low to high
    idxs = jnp.argsort(x)
    x = x[idxs]
    sp = {k: v[idxs] for k, v in sp.items()}

    @jax.jit
    def model(wl: Float = wl):
        S = {}
        zero = jnp.zeros_like(x)

        for key in sp:
            if not key.startswith("wav"):
                port_mode0, port_mode1 = key.split(",")
                port0, mode0 = port_mode0.split("@")
                port1, mode1 = port_mode1.split("@")
                S[(port0, port1)] = jnp.interp(wl, x, sp.get(key, zero))

        return S

    return model


def model_from_csv(
    filepath: Union[PathType, pd.DataFrame],
    xkey: str = "wavelengths",
    xunits: float = 1,
    prefix: str = "s",
) -> Model:
    """Returns a SAX Sparameters Model from a CSV file.

    The SAX Model is a function that returns a SAX SDict interpolated over wavelength.

    Args:
        filepath: CSV Sparameters path or pandas DataFrame.
        wl: wavelength to interpolate (um).
        xkey: key for wavelengths in file.
        xunits: x units in um from the loaded file (um). 1 means 1um.
        prefix: for the sparameters column names in file.
    """
    df = filepath if isinstance(filepath, pd.DataFrame) else pd.read_csv(filepath)
    assert isinstance(df, pd.DataFrame)
    df = df.reset_index()  # maybe there is useful info in the index...
    dic = dict(zip(df.columns, jnp.asarray(df.values.T)))
    keys = list(dic.keys())

    if xkey not in keys:
        raise ValueError(f"{xkey!r} not in {keys}")

    nsparameters = (len(keys) - 1) // 2
    nports = int(nsparameters**0.5)

    x = jnp.asarray(dic[xkey] * xunits)
    wl = jnp.asarray(wl_cband)

    # make sure x is sorted from low to high
    idxs = jnp.argsort(x)
    x = x[idxs]
    dic = {k: v[idxs] for k, v in dic.items()}

    @jax.jit
    def model(wl: Float = wl):
        S = {}
        zero = jnp.zeros_like(x)
        for i in range(1, nports + 1):
            for j in range(1, nports + 1):
                m = jnp.interp(wl, x, dic.get(f"{prefix}{i}{j}m", zero))
                a = jnp.interp(wl, x, dic.get(f"{prefix}{i}{j}a", zero))
                S[f"o{i}", f"o{j}"] = m * jnp.exp(1j * a)

        return S

    return model


def _demo_mmi_lumerical_csv() -> None:
    import matplotlib.pyplot as plt
    from plot_model import plot_model

    filepath = get_sparameters_path_lumerical(gf.c.mmi1x2)
    mmi = model_from_csv(filepath=filepath, xkey="wavelengths")
    plot_model(mmi)
    plt.show()


def model_from_component(component, simulator: Simulator, **kwargs) -> Model:
    """Returns SAX model from lumerical FDTD simulations.

    Args:
        component: to simulate.
        simulator: meep, lumerical or tidy3d.
        kwargs: simulator settings.

    """
    simulators = ["lumerical", "meep", "tidy3d"]

    if simulator == "lumerical":
        filepath = get_sparameters_path_lumerical(component=component, **kwargs)
    elif simulator == "meep":
        filepath = get_sparameters_path_meep(component=component, **kwargs)
    elif simulator == "tidy3d":
        filepath = get_sparameters_path_tidy3d(component=component, **kwargs)
    else:
        raise ValueError(f"{simulator!r} no in {simulators}")
    return model_from_csv(filepath=filepath)


model_from_component_lumerical = gf.partial(model_from_component, simulator="fdtd")


if __name__ == "__main__":
    # mmi1x2 = gf.partial(model_from_component_lumerical, component=gf.components.mmi1x2)
    # mmi2x2 = gf.partial(model_from_component_lumerical, component=gf.components.mmi2x2)

    # grating_coupler_elliptical = model_from_csv(
    #     filepath=sparameters_path / "grating_coupler_ellipti_9d85a0c6_18c08cac.csv"
    # )

    # model_factory = dict(
    #     mmi1x2=mmi1x2, mmi2x2=mmi2x2, grating_coupler_elliptical=grating_coupler_elliptical
    # )
    # from gdsfactory.config import sparameters_path

    import matplotlib.pyplot as plt
    from plot_model import plot_model

    # s = from_csv(filepath=filepath, wavelength=wl_cband, xunits=1e-3)
    # s21 = s[("o1", "o3")]
    # plt.plot(wl_cband, np.abs(s21) ** 2)
    # plt.show()
    # filepath = get_sparameters_path_lumerical(gf.components.mmi1x2)
    # mmi = model_from_csv(filepath=filepath, xkey="wavelengths")
    # plot_model(mmi)
    # plt.show()
    # plot_model(grating_coupler_elliptical)
    # plt.show()
    # model = partial(
    #     from_csv, filepath=filepath, xunits=1, xkey="wavelengths", prefix="s"
    # )
    # sax.plot.plot_model(model)
    # plt.show()
    #  This looks correct
    # coupler_fdtd = gf.partial(
    #     sdict_from_csv,
    #     filepath=gf.config.sparameters_path / "coupler" / "coupler_G224n_L20_S220.csv",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    # this looks wrong
    # coupler_fdtd = model_from_csv(
    #     filepath=sparameters_path / "coupler" / "coupler_G224n_L20_S220.csv",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    # coupler_fdtd = model_from_npz(
    #     filepath=sparameters_path / "coupler" / "coupler_G224n_L20_S220.npz",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    filepath = get_sparameters_path_tidy3d(gf.c.mmi1x2)
    coupler_fdtd = model_from_npz(filepath)
    plot_model(coupler_fdtd)
    plt.show()



"""You can route all component optical ports to a fiber array."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.samples.big_device import big_device


def test_big_device() -> Component:
    component = big_device(nports=10)
    radius = 5.0
    return gf.routing.add_fiber_array(
        component=component, radius=radius, fanout_length=50.0
    )


if __name__ == "__main__":
    c = test_big_device()
    c.show(show_ports=True)



from __future__ import annotations

from typing import Dict, List, Union

import gdsfactory as gf
from gdsfactory.components.straight import straight
from gdsfactory.difftest import difftest
from gdsfactory.port import Port
from gdsfactory.typings import ComponentSpec, Routes


def get_routes_straight(
    ports: Union[List[Port], Dict[str, Port]],
    straight: ComponentSpec = straight,
    **kwargs,
) -> Routes:
    """Returns routes made by 180 degree straights.

    Args:
        ports: List or dict of ports.
        straight: function for straight.
        kwargs: waveguide settings.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_routes_straight")
        pad_array = gf.components.pad_array()
        c1 = c << pad_array
        c2 = c << pad_array
        c2.ymax = -200

        routes = gf.routing.get_routes_straight(ports=c1.get_ports_list(), length=200)
        c.add(routes.references)
        c.plot()

    """
    ports = list(ports.values()) if isinstance(ports, dict) else ports
    straight = straight(**kwargs)
    references = [straight.ref() for _ in ports]
    references = [ref.connect("o1", port) for port, ref in zip(ports, references)]
    ports = [ref.ports["o2"] for ref in references]
    lengths = [straight.info["length"]] * len(ports)
    return Routes(references=references, ports=ports, lengths=lengths)


def test_get_routes_straight(check: bool = True):
    c = gf.Component("get_routes_straight")
    pad_array = gf.components.pad_array()
    c1 = c << pad_array
    c2 = c << pad_array
    c2.ymax = -200

    routes = get_routes_straight(ports=c1.get_ports_list(), length=200)
    c.add(routes.references)
    if check:
        difftest(c)
    return c


if __name__ == "__main__":
    c = test_get_routes_straight(False)
    c.show(show_ports=True)



from __future__ import annotations

import itertools as it
from typing import Optional

import numpy as np
from shapely import geometry
from shapely.geometry.polygon import Polygon

from gdsfactory.geometry.functions import polygon_grow
from gdsfactory.typings import Coordinates

DEG2RAD = np.pi / 180
RAD2DEG = 1.0 / DEG2RAD
# from matplotlib import pyplot as plt


def pixelate_path(
    pts: Coordinates,
    pixel_size: float = 0.55,
    snap_res: float = 0.05,
    middle_offset: float = 0.5,
    theta_start: float = 0,
    theta_end: float = 90,
) -> Coordinates:
    """From a path add one pixel per point on the path.

    Args:
        pts: points.
        pixel_size: in um.
        snap_res: snap resolution.
        middle_offset: in um.
        theta_start: in degrees.
        theta_end: in degrees.

    """
    thetas0 = [
        np.arctan2(y1 - y0, x1 - x0) for (x0, y0), (x1, y1) in zip(pts[:-1], pts[1:])
    ]
    thetas0 += [theta_end * DEG2RAD]
    thetas0[0] = theta_start * DEG2RAD

    thetas1 = [
        np.arctan2(x1 - x0, y1 - y0) for (x0, y0), (x1, y1) in zip(pts[:-1], pts[1:])
    ]
    thetas1[0] = (theta_start - 180) * DEG2RAD
    thetas1 += [(theta_end - 180) * DEG2RAD]

    thetas_deg0 = np.abs(np.array(thetas0) * RAD2DEG) % 90
    thetas_deg1 = np.abs(np.array(thetas1) * RAD2DEG) % 90

    thetas_deg = thetas_deg1

    slice = np.where(thetas_deg0 < 45)
    thetas_deg[slice] = thetas_deg0[slice]

    scalings = np.cos(
        abs(thetas_deg) * DEG2RAD
    )  # + middle_offset * (1 - np.cos(abs(thetas_deg) * DEG2RAD) )

    def _snap(x: float) -> float:
        return round(int(x / snap_res), 0) * snap_res

    def _gen_pixel(p, a):
        x0, y0 = p
        pix = [(x0 + a, y0 - a), (x0 + a, y0 + a), (x0 - a, y0 + a), (x0 - a, y0 - a)]
        pix = [(_snap(_x), _snap(_y)) for _x, _y in pix]
        return pix

    a = pixel_size / 2
    return [_gen_pixel(p, a * s) for p, s in zip(pts, scalings)]


def points_to_shapely(pts: Coordinates) -> Polygon:
    return Polygon(pts)


def _snap_to_resolution(a, snap_res):
    return np.round(a / snap_res, 0) * snap_res


def _pixelate(
    pts: Coordinates,
    N: int = 100,
    margin: float = 0.4,
    margin_x: Optional[float] = None,
    margin_y: Optional[float] = None,
    nb_pixels_x: Optional[int] = None,
    nb_pixels_y: Optional[int] = None,
    min_pixel_size: float = 0.4,
    snap_res: float = 0.05,
) -> Coordinates:
    """Pixelates a shape (as 2d array) onto an NxN grid.

    Arguments:
        pts: The 2D array to be pixelated.
        N: The number of pixels on an edge of the grid.

    Returns:
        A list of pixel bounding boxes

    """
    shape = points_to_shapely(pts)  # convert to shapely
    if not shape:
        return []

    if margin_x is None:
        margin_x = margin

    if margin_y is None:
        margin_y = margin

    if nb_pixels_x is None:
        nb_pixels_x = N

    if nb_pixels_y is None:
        nb_pixels_y = N

    west, south, east, north = shape.bounds

    width = east - west + 2 * margin_x
    height = north - south + 2 * margin_y

    if min_pixel_size is not None:
        max_nb_pixels_x = int(np.ceil(width / min_pixel_size))
        nb_pixels_x = min(nb_pixels_x, max_nb_pixels_x)

    if min_pixel_size is not None:
        max_nb_pixels_y = int(np.ceil(height / min_pixel_size))
        nb_pixels_y = min(nb_pixels_y, max_nb_pixels_y)

    w = width / nb_pixels_x + snap_res
    h = height / nb_pixels_y + snap_res

    ax = margin_x
    ay = margin_y

    xs = np.linspace(west + w / 2 - ax, east - w / 2 + ax, nb_pixels_x)
    ys = np.linspace(south + h / 2 - ay, north - h / 2 + ay, nb_pixels_y)
    xs = _snap_to_resolution(xs, snap_res)
    ys = _snap_to_resolution(ys, snap_res)
    grid = it.product(xs, ys)

    pixels = []

    for x, y in grid:
        _w, _s, _e, _n = x - w / 2, y - h / 2, x + w / 2, y + h / 2
        newpix = geometry.box(_w, _s, _e, _n)
        if shape.intersects(newpix):
            r = (_w, _s, _e, _n)

            pixels.append(r)
    return pixels


def rect_to_coords(r):
    w, s, e, n = r
    return [(w, s), (w, n), (e, n), (e, s)]


def pixelate(pts, N=100, margin=0.4, **kwargs):
    """Pixelate shape defined by points Return rectangles [Rect1, Rect2, ...] \
    ready to go in the quad tree."""
    pixels = _pixelate(pts, N=N, margin=margin, **kwargs)
    return [rect_to_coords(pixel) for pixel in pixels]


def gen_pixels_op_blocking(pts, snap_res=0.05, margin=1.0, min_pixel_size=0.4):
    op_block_pts = polygon_grow(pts, margin)
    return pixelate(
        op_block_pts, min_pixel_size=min_pixel_size, N=100, snap_res=snap_res
    )


def gen_op_blocking(pts, snap_res=0.05, margin=0.3):
    return polygon_grow(pts, margin)


if __name__ == "__main__":
    import numpy as np

    pts = [(x, x**2) for x in np.linspace(0, 1, 5)]
    c = pixelate(pts)
    print(c)



from __future__ import annotations
from collections.abc import Iterable
from typing import Any, Dict, List
from gdsfactory.component import Component
import gdsfactory as gf
from gdsfactory.get_netlist import get_netlist_recursive


def get_netlist_flat(
    component: Component,
    **kwargs,
) -> Dict[str, Any]:
    """Parses a recursive netlist for a component as if it was a single netlist \
            with its lowest-level instances.

    Procedure:
        - Recursively parse the recursive dict to generate a unique list of all instances **even if they are reused**
          Each entry is formatted as [(netlist, instance),...], and defines a unique flat_name = {top}{hierarchy_delimiter}{instance}{hierarchy_delimiter}{instance}{hierarchy_delimiter}...
        - Populate the flat netlist dict with similar entries as regular gdsfactory netlists:
            - instances: component, info, and settings from the original netlist, but keyed with flat_name
            - placements: For instance corresponding to each flat_name, accumulate placements across the hierarchy
            - ports and connections: each unique port of each flat_name instance is uniquely named as flat_name,port_name
                                     for each flat_name,port_name, starting at lowest hierarchy level:
                                        - list connections at that level (finding the proper other flat_name,port_name if non-leaf level)
                                        - if possible, maps the port to a port of that instance's netlist, and repeat at a higher level
                                        - if the top level is reached, assign that flat_name,port_name to a top-level component port instead
                    the returned ports dict has top level component portname: flat_name,port_name mappings
                    the returned connections dict is sorted and flattened into a minimal (flat_name,port_name)_1: [(flat_name,port_name)] key: value pairs
                        If allow_multiple flag in get_netlist is True, the values will be lists (to support multiple connections)
            - name: top_level_component name

    Args:
        component: to extract flat netlist.

    Keyword Args:
        component_suffix: suffix to append to each component name.
            useful if to save and reload a back-annotated netlist.
        get_netlist_func: function to extract individual netlists.
        full_settings: True returns all, false changed settings.
        tolerance: tolerance in nm to consider two ports connected.
        exclude_port_types: optional list of port types to exclude from netlisting.
        get_instance_name: function to get instance name.

    Returns:
        instances: Dict of instance name and settings.
        connections: Dict of Instance1Name,portName: Instance2Name,portName.
        placements: Dict of instance names and placements (x, y, rotation).
        port: Dict portName: ComponentName,port.
        name: name of component.
        warnings: warning messages (disconnected pins).
    """
    recursive_netlist = get_netlist_recursive(component, **kwargs)
    top_level_name = component.name
    hierarchical_instances = _flatten_hierarchy(top_level_name, recursive_netlist)
    connections = {}
    ports = {}
    placements = {}
    instances = {}
    for hierarchical_instance in hierarchical_instances:
        c, p = _map_connections_ports(
            hierarchical_instance, top_level_name, recursive_netlist
        )
        for key, value in c.items():
            if len(value) != 0:
                connections[key] = value
        for key, value in p.items():
            if len(value) != 0:
                ports[key] = value
        placements.update(
            _accumulate_placements(hierarchical_instance, recursive_netlist)
        )
        instances.update(_get_instance_info(hierarchical_instance, recursive_netlist))

    return {
        "connections": connections,
        "placements": placements,
        "instances": instances,
        "ports": ports,
        "name": recursive_netlist[top_level_name]["name"],
    }


def _flat_name(
    hierarchy,
    hierarchy_delimiter: str = "~",
):
    """Returns a unique name for the instance from its hierarchy."""
    name = ""
    for _, instance in hierarchy[:-1]:
        name += instance
        name += hierarchy_delimiter
    name += hierarchy[-1][1]
    return name


def _get_instance_info(hierarchy, recursive_netlist):
    """Returns instance info from its instance data."""
    return {
        _flat_name(hierarchy): recursive_netlist[hierarchy[-2][0]]["instances"][
            hierarchy[-1][1]
        ]
    }


def _get_leaf(
    instance_port,
    netlist,
    flat_name,
    all_netlists,
    hierarchy_delimiter: str = "~",
):
    """Given a instance, port and its netlist, maps ports down to the lowest hierarchy level."""
    instance, port = instance_port.split(",")
    netlist = all_netlists[netlist]["instances"][instance]["component"]
    found_lower_level = netlist in all_netlists.keys()
    if found_lower_level:
        instance_port = all_netlists[netlist]["ports"][port]
        flat_name = f"{flat_name}{hierarchy_delimiter}{instance}"
        found_lower_level, flat_name = _get_leaf(
            instance_port, netlist, flat_name, all_netlists
        )
    else:
        found_lower_level = False
        flat_name = f"{flat_name}{hierarchy_delimiter}{instance_port}"
    return found_lower_level, flat_name


def _lateral_map(
    local_leaf_port: str,
    all_netlists: Dict[str, Any],
    higher_component: str,
    level: int,
    hierarchy,
    hierarchy_delimiter: str = "~",
):
    """Returns connected ports at this hierarchical level."""
    lateral_equivalencies = []
    for port1, port2 in all_netlists[higher_component]["connections"].items():
        flat_name_prefix = _flat_name(hierarchy[: len(hierarchy) - level - 1])
        if local_leaf_port == port1 and local_leaf_port != port2:
            lateral_equivalencies.append(
                _get_leaf(port2, higher_component, flat_name_prefix, all_netlists)[1]
            )
        elif local_leaf_port == port2 and local_leaf_port != port1:
            lateral_equivalencies.append(
                _get_leaf(port1, higher_component, flat_name_prefix, all_netlists)[1]
            )
    return lateral_equivalencies


def _map_connections_ports(
    hierarchy,
    top_name,
    all_netlists,
    hierarchy_delimiter: str = "~",
):
    """Returns all nodes of interest across the flat recursive netlist."""
    connections = {}
    ports = {}

    # Starting point is ports of the leaf instance
    leaf_instance = hierarchy[-1][1]
    leaf_instance_ports = list(gf.get_component(hierarchy[-1][0]).ports.keys())

    for leaf_portname in leaf_instance_ports:
        current_connections = []
        current_ports = []
        local_leaf_port = f"{leaf_instance},{leaf_portname}"
        for level, (higher_component, higher_instance) in enumerate(
            hierarchy[:-1][::-1]
        ):
            # Identify connected node (with flattened names)
            current_connections.extend(
                _lateral_map(
                    local_leaf_port,
                    all_netlists,
                    higher_component,
                    level,
                    hierarchy,
                    hierarchy_delimiter,
                )
            )
            # Traverse up to higher level
            found_higher_level = False
            for portname, port in all_netlists[higher_component]["ports"].items():
                if local_leaf_port == port:
                    local_leaf_port = f"{higher_instance},{portname}"
                    found_higher_level = True
                    # If there is a port at top level, map to that as well
                    if higher_instance == top_name:
                        current_ports.append(local_leaf_port)
            if not found_higher_level:
                break
        flat_leaf_port = f"{_flat_name(hierarchy)},{leaf_portname}"
        connections[flat_leaf_port] = current_connections
        ports[flat_leaf_port] = current_ports

    return connections, ports


def _accumulate_placements(
    hierarchy,
    all_netlists,
):
    """Iterate through hierarchy tuples, accumulating placement information."""
    placements = {key: 0 for key in ["x", "y", "mirror", "rotation"]}
    for (higher_component, _higher_instance), (_lower_component, lower_instance) in zip(
        hierarchy[:-1], hierarchy[1:]
    ):
        for key in ["x", "y", "mirror", "rotation"]:
            placements[key] += all_netlists[higher_component]["placements"][
                lower_instance
            ][key]

    return {_flat_name(hierarchy): placements}


def _flatten_str_list(xs):
    """Flatten nested list of strings to list of strings."""
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from _flatten_str_list(x)
        else:
            yield x


def _flatten_hierarchy(
    netlist_name: str,
    all_netlists: List[Dict[str, any]],
    hierarchy_delimiter: str = "~",
    component_instance_delimiter: str = ";",
):
    """Converts _flatten_netlist output str's to list of hierarchical (component, instance) tuples."""

    instance_dict = _flatten_hierarchy_recurse(
        netlist_name=netlist_name,
        all_netlists=all_netlists,
        hierarchy_delimiter=hierarchy_delimiter,
        component_instance_delimiter=component_instance_delimiter,
    )

    hierarchies_lists = []
    for hierarchy in _flatten_str_list(list(instance_dict.keys())):
        hierarchy_list = []
        levels = hierarchy.split(hierarchy_delimiter)
        for level in levels:
            components, instance = level.split(component_instance_delimiter)
            hierarchy_list.append((components, instance))
        hierarchies_lists.append(hierarchy_list)

    return hierarchies_lists


def _flatten_hierarchy_recurse(
    netlist_name: str,
    all_netlists: List[Dict[str, any]],
    hierarchy: str = None,
    hierarchy_delimiter: str = "~",
    component_instance_delimiter: str = ";",
) -> Dict[str, Any]:
    """Flattens the provided recursive netlist by recursively updating a list.

    Args:
        netlist_name: netlist entry to flatten
        all_netlists: list of all possible netlists (output of get_netlist_recursive)
        hierarchy: str to append to netlist_name
        hierarchy_delimiter: str to separate hierarchy levels in flattened keys
        component_instance_delimiter: str to separate instance name from component name for netlist reconstruction

    Returns:
        str of ...{hierarchy_delimiter}{component}{component_instance_delimiter}{instance}{hierarchy_delimiter} used to flatten the netlist
    """
    instance_dict = {}
    if hierarchy is None:
        top_component = all_netlists[netlist_name]["name"]
        hierarchy = f"{top_component}{component_instance_delimiter}{netlist_name}"
    for instance_name, instance in all_netlists[netlist_name]["instances"].items():
        component_name = instance["component"]
        hierarchy_str = f"{hierarchy}{hierarchy_delimiter}{component_name}{component_instance_delimiter}{instance_name}"
        if component_name not in all_netlists.keys():
            instance_dict[hierarchy_str] = True  # Done for this leaf
        else:
            instance_dict.update(
                _flatten_hierarchy_recurse(
                    component_name,
                    all_netlists,
                    hierarchy_str,
                )
            )

    return instance_dict


if __name__ == "__main__":
    import pprint

    coupler_lengths = [10, 20, 30, 40]
    coupler_gaps = [0.1, 0.2, 0.4, 0.5]
    delta_lengths = [10, 100, 200]

    c = gf.components.mzi_lattice(
        coupler_lengths=coupler_lengths,
        coupler_gaps=coupler_gaps,
        delta_lengths=delta_lengths,
    )
    flat_netlist = get_netlist_flat(c)
    print("")
    print("FLAT NETLIST")
    print("")
    pprint.pprint(flat_netlist)

    # """
    # Testing electrical netlist w/ identical component references
    # """
    # # Define compound component
    # series_resistors = gf.Component("seriesResistors")
    # rseries1 = series_resistors << gf.get_component(
    #     gf.components.resistance_sheet, width=20, ohms_per_square=20
    # )
    # rseries2 = series_resistors << gf.get_component(
    #     gf.components.resistance_sheet, width=20, ohms_per_square=20
    # )
    # rseries1.connect("pad2", rseries2.ports["pad1"])
    # series_resistors.add_port("pad1", port=rseries1.ports["pad1"])
    # series_resistors.add_port("pad2", port=rseries2.ports["pad2"])

    # # Increase hierarchy levels more
    # double_series_resistors = gf.Component("double_seriesResistors")
    # rseries1 = double_series_resistors << gf.get_component(series_resistors)
    # rseries2 = double_series_resistors << gf.get_component(series_resistors)
    # rseries1.connect("pad2", rseries2.ports["pad1"])
    # double_series_resistors.add_port("pad1", port=rseries1.ports["pad1"])
    # double_series_resistors.add_port("pad2", port=rseries2.ports["pad2"])

    # # Define top-level component
    # vdiv = gf.Component("voltageDivider")
    # r1 = vdiv << double_series_resistors
    # r2 = vdiv << series_resistors
    # r3 = (
    #     vdiv
    #     << gf.get_component(
    #         gf.components.resistance_sheet, width=20, ohms_per_square=20
    #     ).rotate()
    # )
    # r4 = vdiv << gf.get_component(
    #     gf.components.resistance_sheet, width=20, ohms_per_square=20
    # )

    # r1.connect("pad2", r2.ports["pad1"])
    # r3.connect("pad1", r2.ports["pad1"], preserve_orientation=True)
    # r4.connect("pad1", r3.ports["pad2"], preserve_orientation=True)

    # vdiv.add_port("gnd1", port=r2.ports["pad2"])
    # vdiv.add_port("gnd2", port=r4.ports["pad2"])
    # vdiv.add_port("vsig", port=r1.ports["pad1"])
    # vdiv.show(show_ports=True)

    # recursive_netlist = get_netlist_recursive(vdiv, allow_multiple=True)
    # import pprint

    # print("RECURSIVE NETLIST")
    # print("")
    # pprint.pprint(recursive_netlist)
    # print("")
    # print("OPERATION")
    # print("")

    # flat_netlist = get_netlist_flat(vdiv, allow_multiple=True)
    # print("")
    # print("FLAT NETLIST")
    # print("")
    # pprint.pprint(flat_netlist)



"""CD SEM structures."""
from __future__ import annotations

from functools import partial
from typing import Optional, Tuple

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.typings import ComponentSpec, CrossSectionSpec

text_rectangular_mini = partial(text_rectangular, size=1)

LINE_LENGTH = 420.0


@cell
def cdsem_straight(
    widths: Tuple[float, ...] = (0.4, 0.45, 0.5, 0.6, 0.8, 1.0),
    length: float = LINE_LENGTH,
    cross_section: CrossSectionSpec = "strip",
    text: Optional[ComponentSpec] = text_rectangular_mini,
    spacing: float = 5,
) -> Component:
    """Returns straight waveguide lines width sweep.

    Args:
        widths: for the sweep.
        length: for the line.
        cross_section: for the lines.
        text: optional text for labels.
        spacing: edge to edge spacing.
    """
    c = Component()

    ymin = 0

    for width in widths:
        line = c << straight(length=length, cross_section=cross_section, width=width)
        line.ymin = ymin
        if text:
            t = c << text(str(int(width * 1e3)))
            t.xmin = line.xmax + 5
            t.y = ymin

        ymin += width + spacing

    return c


if __name__ == "__main__":
    c = cdsem_straight()
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.components.straight_heater_doped_rib()
    # c = gf.components.mzi_phase_shifter_top_heater_metal()
    # c = gf.routing.add_fiber_array(c)
    # c = gf.routing.add_electrical_pads_top(c)
    c.show()

    # scene = c.to_3d()
    # scene.show()



from __future__ import annotations

from typing import Any, List, Tuple

import numpy as np
from scipy.constants import speed_of_light
from simphony import Model
from simphony.pins import Pin, PinList
from simphony.tools import interpolate

import gdsfactory as gf
import gdsfactory.simulation.lumerical as sim
from gdsfactory.component import Component


class GDSFactorySimphonyWrapper(Model):
    """Take a GDSFactory component and convert it into a Simphony Model object."""

    def __init__(
        self,
        name: str = "",
        *,
        component: Component,
        dirpath=gf.PATH.sparameters,
        **kwargs,
    ) -> None:
        """Take a GDSFactory component and convert it into a Simphony Model object.

        Args:
            name: name of the model.
            component: component factory or instance.
            dirpath: sparameters directory.
            kwargs: settings.

        """
        pin_names, self.f, self.s = self._model_from_gdsfactory(
            component=component, dirpath=dirpath, **kwargs
        )

        pins = PinList(
            [Pin(component=self, name=pin_names[i]) for i in range(len(pin_names))]
        )

        freq_range = self.f[0], self.f[-1]

        self.wavelengths = speed_of_light / np.array(self.f)

        super().__init__(name, freq_range=freq_range, pins=pins)

    def s_parameters(self, freqs: np.array) -> np.ndarray:
        return interpolate(freqs, self.f, self.s)

    def _model_from_gdsfactory(
        self, component: Component, dirpath=gf.PATH.sparameters, **kwargs
    ) -> Tuple[List[str], Any, np.ndarray]:
        """Return simphony model from gdsfactory Component Sparameters.

        Args:
            component: component factory or instance.
            dirpath: sparameters directory.
            kwargs: settings.

        """
        kwargs.pop("function_name", "")
        kwargs.pop("module", "")
        component = gf.call_if_func(component, **kwargs)

        return sim.read_sparameters_lumerical(component=component, dirpath=dirpath)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    c = GDSFactorySimphonyWrapper(component=gf.c.mmi2x2())
    # c = GDSFactorySimphonyWrapper(component=gf.c.mmi2x2())
    # c = GDSFactorySimphonyWrapper(component=gf.c.bend_euler())
    # wav = np.linspace(1520, 1570, 1024) * 1e-9
    # f = speed_of_light / wav
    # s = c.s_parameters(freqs=f)

    wav = c.wavelengths
    s = c.s
    plt.plot(wav * 1e9, np.abs(s[:, 1] ** 2))
    plt.show()



import copy
from pathlib import Path
from typing import Dict, Optional, Union

import jax.numpy as jnp
from tqdm.contrib.itertools import product

from gdsfactory.simulation.sax.interpolators import nd_nd_interpolation
from gdsfactory.simulation.sax.mlp import mlp_regression
from gdsfactory.simulation.sax.parameter import (
    LayerStackThickness,
    NamedParameter,
    LithoParameter,
)
from gdsfactory.technology import LayerStack
from gdsfactory.typings import PortSymmetries
import ray


class Model:
    def __init__(
        self,
        trainable_component: callable,
        layerstack: LayerStack,
        trainable_parameters: Dict[
            str, Union[LayerStackThickness, NamedParameter]
        ] = None,
        non_trainable_parameters: Dict[
            str, Union[LayerStackThickness, NamedParameter]
        ] = None,
        simulation_settings: Optional[Dict[str, Union[float, str, int, Path]]] = None,
        num_modes: int = 2,
        port_symmetries: Optional[PortSymmetries] = None,
        address: str = None,
        dashboard_port: int = 8265,
        num_cpus: int = None,
        num_cpus_per_task: int = 1,
        # num_gpus_per_task: int = 0,
        restart_cluster: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Utility class which simplifies the execution of simulations to build compact models for circuit simulations.
        Aims to be agnostic to the specific simulator being used.

        It contains shared utilities for the different types of Models:
            - Optional simulation hyperparameter tuning routine (TODO)
            - Looping over training variables to generate training examples.
            - Consistent formatting of training examples for model building.
            - Interface with Ray clusters for distributed processing (even locally, this is beneficial)

        Other functionality such as
            - Simulation setup, execution, caching/loading
            - Assembly of simulation results into S-parameters
        is solver-dependent, and hence resides in child classes.

        TODO:
            - more consistent ordering of input/output data
            - more JAX, less pure Python
            - reuse Ray cluster across different model instances for simultaneous training

        Attributes:
            trainable_component: callable wrapping component associated with model
            layerstack: complete layerstack associated with model
            trainable parameters: parameters that are model features that need to be learned (e.g. width, layer thicknesses, wavelength, materials properties)
            non_trainable_parameters: parameters for input_dict whose effect on S-parameters is known and does not need to be trained (e.g. length for a waveguide)
            simulation_settings: simulation parameters that can impact model quality (e.g. resolution, domain size). Set their converged property to True to skip convergence tuning.
            sim_settings: other simulation settings the solver can use
            num_modes: number of modes to consider for each port in S-parameter calculation.
            port_symmetries: as defined in FDTD solvers. Dict establishing equivalency between S-parameters.
                e.g. {"o1@0,o1@0": ["o2@0,o2@0", "o3@0,o3@0", "o4@0,o4@0"]} means that S22 = S33 = S44 are the same as S11,
                and hence the output vector for model training will only contain the key S11 to reduce the number of variables to fit.
            address: of the Ray cluster to connect to. Defaults to finding a local running instance.
            dashboard_port: IP address of the dashboard to monitor the cluster
            num_cpus: available to the cluster (if not autoscaling)
            num_cpus_per_task: number of CPUs to assign to each task
            num_gpus_per_task: number of GPUs to assign to each task
            restart_cluster: if instantiating multiple models in the same Python session, whether to restart the cluster.
        """
        self.trainable_component = trainable_component
        self.layerstack = layerstack
        self.trainable_parameters = trainable_parameters or {}
        self.non_trainable_parameters = non_trainable_parameters or {}
        self.simulation_settings = simulation_settings or {}
        self.num_modes = num_modes

        # Extract inputs and outputs vector size data
        self.size_inputs = len(self.trainable_parameters)
        self.num_ports = len(
            trainable_component(self.get_nominal_dict()).get_ports_list(
                port_type="optical"
            )
        )

        # Extract input and output vector label data
        self.input_vector_labels = list(self.trainable_parameters) + list(
            self.non_trainable_parameters.keys()
        )

        # self.size_outputs = self.num_ports * self.num_modes
        self.port_symmetries = port_symmetries

        # Cluster resources
        self.num_cpus_per_task = num_cpus_per_task
        # self.num_gpus_per_task = num_gpus_per_task
        if restart_cluster and ray.is_initialized():
            ray.shutdown()
        if not ray.is_initialized():
            ray.init(dashboard_port=dashboard_port, num_cpus=num_cpus)

    """
    PARAMETERS
    """

    def get_nominal_dict(self):
        """Return input_dict of nominal parameter values."""
        return {
            name: parameter.nominal_value
            for name, parameter in self.trainable_parameters.items()
        }

    def get_random_dict(self):
        """Return input_dict of randomly sampled parameter values."""
        return {
            name: parameter.sample()
            for name, parameter in self.trainable_parameters.items()
        }

    def parse_input_dict(self, input_dict):
        """Separates between LayerStackThickness, NamedParameter, and LithoParameter inputs.

        Args:
            input_dict: key needs to match the keys in self.trainable_parameters
                values are the new values to assign to these parameters.
        """
        param_dict = {}
        layerstack_param_dict = {}
        litho_param_dict = {}
        for key, value in input_dict.items():
            if type(self.trainable_parameters[key]) is NamedParameter:
                param_dict[key] = value
            elif type(self.trainable_parameters[key]) is LayerStackThickness:
                layerstack_param_dict[key] = value
            elif type(self.trainable_parameters[key]) is LithoParameter:
                litho_param_dict[key] = value
        return param_dict, layerstack_param_dict, litho_param_dict

    def perturb_layerstack(self, layerstack_param_dict):
        """Returns a temporary LayerStack with a new thickness value for the (current) LayerStackThickness objects in layerstack_param_dict.

        Args:
            layerstack_param_dict: key needs to match a key in self.trainable_parameters having for value a LayerStackThickness object
                                    value is the thickness to assign to this parameter
        """
        perturbed_layerstack = copy.deepcopy(self.layerstack)
        for key, thickness in layerstack_param_dict.items():
            LayerStackThickness_obj = self.trainable_parameters[key]
            perturbed_layerstack.layers[
                LayerStackThickness_obj.layername
            ].thickness = thickness
        return perturbed_layerstack

    def perturb_geometry(self, current_component, litho_param_dict):
        """Returns a temporary Component on which all the morphological operations contained in the current litho_param_dict have been applied.

        Args:
            current_component: the current component, with params_dict already applied
            litho_param_dict: key needs to match a key in self.trainable_parameters having for value a LithoParameter object
                                    value is the input to the LithoParameter transformation attribute
        """
        for key, value in litho_param_dict.items():
            lithoParameter_obj = self.trainable_parameters[key]
            current_component = lithoParameter_obj.get_transformation(
                current_component, value
            )
        return current_component

    """
    QUEUING
    """

    def define_output_vector_labels(self):
        """Uses number of component ports, number of modes solved for, and port_symmetries to define smallest output vector."""
        output_vector_labels_iter = product(
            range(1, self.num_ports + 1),
            range(self.num_modes),
            range(1, self.num_ports + 1),
            range(self.num_modes),
        )
        output_vector_labels = []
        for output_label in output_vector_labels_iter:
            output_key1 = f"o{output_label[0]}@{output_label[1]}"
            output_key2 = f"o{output_label[2]}@{output_label[3]}"
            if output_key1 != output_key2:
                output_key = f"{output_key1},{output_key2}"
                output_vector_labels.append(output_key)

        if self.port_symmetries:
            for value_list in self.port_symmetries.values():
                for value in value_list:
                    output_vector_labels.remove(value)

        return output_vector_labels

    def arange_inputs(self, type="arange"):
        """Prepares input vectors for spanning simulation space.

        Arguments:
            type: str, arange or corners. Defines the iterator function to use with parameter objects.

        Returns:
            ranges_dict: dict, with keys parameter names and values arrays of containing all unique parameter values.
        """
        if type == "arange":
            ranges_dict = {
                name: parameter.arange()
                for name, parameter in self.trainable_parameters.items()
            }
        elif type == "corners":
            ranges_dict = {
                name: parameter.corners()
                for name, parameter in self.trainable_parameters.items()
            }
        else:
            raise ValueError("Type should be arange or corners.")

        return ranges_dict

    def get_output_from_inputs(self, labels, values, remote_function):
        """Get one output vector from a set of inputs.

        How to map parameters to simulation inputs depends on the target simulator.

        Arguments:
            labels: keys of the parameters
            values: values of the parameters
            remote_function: ray remote function object to use for the simulation

        Returns:
            remote function ID for delayed execution
            the remote function returns new_inputs, output_vectors
            new_inputs is an array containing derived inputs (e.g. wavelength in FDTD broadband simulation)
            output_vectors is a vector of reals representing the output to model (neff, s-params, etc.)
        """
        return NotImplementedError

    def get_all_inputs_outputs(self, type="arange"):
        """Get all outputs given all sets of inputs.

        get_output_from_inputs and remote_function are defined in the child class.
        """
        # Define possible parameter values
        ranges_dict = self.arange_inputs(type=type)
        # For all combinations of parameter values
        input_ids = list(product(*ranges_dict.values()))
        output_ids = [
            self.get_output_from_inputs(
                ranges_dict.keys(), values, self.remote_function
            )
            for values in product(*ranges_dict.values())
        ]
        # Execute the jobs
        results = ray.get(output_ids)

        # Parse the outputs into input and output vectors
        input_vectors = []
        output_vectors = []
        for input_example, output_example in zip(
            input_ids, results
        ):  # TODO no for loops!
            input_vector = list(input_example)
            input_vector.extend(output_example[0])
            input_vectors.append(input_vector)
            output_vectors.append(output_example[1])

        return (
            jnp.array(input_vectors),
            jnp.array(output_vectors),
        )

    """
    MODELS
    """

    def set_nd_nd_interp(self):
        """Returns ND-ND interpolator.

        Returns:
            self.inference: [callable giving an output_vector given an input_vector]
            list is of length 2*output_vector length, first output_vector entries are real, second imaginary
        """
        input_vectors, output_vectors = self.get_all_inputs_outputs()
        self.inference = nd_nd_interpolation(input_vectors, output_vectors)

    def set_mlp_interp(self):
        """Returns multilayer perceptron interpolator.

        Returns:
            self.inference: [callable giving an output_vector given an input_vector]
            list is of length 2*output_vector length, first output_vector entries are real, second imaginary
        """
        input_vectors, output_vectors = self.get_all_inputs_outputs()
        self.inference = mlp_regression(input_vectors, output_vectors)

    def input_dict_to_input_vector(self, input_dict):
        """Convert an input_dict with trainable and non-trainable parameters to an input vector on which inference can be performed.

        Find faster way!
        """
        return_array = [
            value
            for key, value in input_dict.items()
            if key in self.trainable_parameters.keys()
        ]
        return jnp.array(return_array)

    def validate(self, num_samples=1):
        """Validates the model by calculating random points within the interpolation range, and comparing to predictions."""
        calculated_outputs_ids = []
        inferred_outputs = []
        validation_inputs = []
        for _index in range(num_samples):
            validation_inputs_local = self.get_random_dict()
            validation_inputs.append(validation_inputs_local)
            calculated_outputs_ids.append(
                self.get_output_from_inputs(
                    validation_inputs_local.keys(),
                    validation_inputs_local.values(),
                    self.remote_function,
                )
            )
            inferred_outputs_local = [
                self.inference[mode](validation_inputs_local.values())
                for mode in range(self.num_modes)
            ]
            inferred_outputs.append(inferred_outputs_local)

        # Execute the jobs
        calculated_outputs_results = ray.get(calculated_outputs_ids)

        # Parse the outputs into input and output vectors
        calculated_outputs = [
            calculated_output_vector[1]
            for calculated_output_vector in calculated_outputs_results
        ]
        return validation_inputs, calculated_outputs, inferred_outputs



from __future__ import annotations

import numpy as np

from gdsfactory.simulation.modes.find_mode_dispersion import find_mode_dispersion


def test_find_modes_waveguide_dispersion() -> None:
    modes = find_mode_dispersion(wg_width=0.45, resolution=20, cache=None)
    m1 = modes

    # print(f"neff1 = {m1.neff}")
    # print(f"ng1 = {m1.ng}")

    # neff1 = 2.3948
    # ng1 = 4.23194

    neff1 = 2.362907833437435
    ng1 = 4.202169359808116

    assert np.isclose(m1.neff, neff1, rtol=1e-2), (m1.neff, neff1)
    assert np.isclose(m1.ng, ng1, rtol=1e-2), (m1.ng, ng1)


if __name__ == "__main__":
    test_find_modes_waveguide_dispersion()



from __future__ import annotations

import typing
from gdsfactory.generic_tech.layer_map import GenericLayerMap as LayerMap
from gdsfactory.generic_tech.layer_stack import LAYER_STACK
from gdsfactory.technology import LayerViews

if typing.TYPE_CHECKING:
    from gdsfactory.pdk import Pdk

LAYER = LayerMap()

PORT_MARKER_LAYER_TO_TYPE = {
    LAYER.PORT: "optical",
    LAYER.PORTE: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_LAYER_TO_TYPE = {
    LAYER.WG: "optical",
    LAYER.WGN: "optical",
    LAYER.SLAB150: "optical",
    LAYER.M1: "dc",
    LAYER.M2: "dc",
    LAYER.M3: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_TYPE_TO_MARKER_LAYER = {v: k for k, v in PORT_MARKER_LAYER_TO_TYPE.items()}

LAYER_TRANSITIONS = {
    LAYER.WG: "taper",
    LAYER.M3: "taper",
    # (LAYER.)
}


def get_generic_pdk() -> Pdk:
    from gdsfactory.components import cells
    from gdsfactory.config import sparameters_path, PATH
    from gdsfactory.cross_section import cross_sections
    from gdsfactory.pdk import Pdk, constants

    LAYER_VIEWS = LayerViews(filepath=PATH.klayout_yaml)

    return Pdk(
        name="generic",
        cells=cells,
        cross_sections=cross_sections,
        layers=LAYER.dict(),
        layer_stack=LAYER_STACK,
        layer_views=LAYER_VIEWS,
        layer_transitions=LAYER_TRANSITIONS,
        sparameters_path=sparameters_path,
        constants=constants,
    )


def write_lyp() -> None:
    """Write from YAML to Klayout lyp file"""
    from gdsfactory.config import PATH

    LAYER_VIEWS = LayerViews(filepath=PATH.klayout_yaml)
    LAYER_VIEWS.to_lyp(PATH.klayout_lyp)


def write_yaml() -> None:
    """Write from YAML to Klayout lyp file"""
    from gdsfactory.config import PATH

    LAYER_VIEWS = LayerViews(filepath=PATH.klayout_lyp)
    LAYER_VIEWS.to_yaml(PATH.generic_tech / "layers2.yaml")


if __name__ == "__main__":
    # pdk = get_generic_pdk()
    # pdk.layer_views.to_yaml('layer_views2.yaml')
    # print(pdk.name)
    write_lyp()
    # write_yaml()



"""Technology settings."""
from __future__ import annotations

import pathlib
from typing import Tuple, Union

from pydantic import BaseModel

from gdsfactory.technology import LayerLevel, LayerStack

module_path = pathlib.Path(__file__).parent.absolute()
Layer = Tuple[int, int]
LayerSpec = Union[int, Layer, str, None]
nm = 1e-3


class LayerMap(BaseModel):
    """Generic layermap based on book.

    Lukas Chrostowski, Michael Hochberg, "Silicon Photonics Design",
    Cambridge University Press 2015, page 353

    You will need to create a new LayerMap with your specific foundry layers.
    """

    WG: Layer = (1, 0)
    WGCLAD: Layer = (111, 0)
    SLAB150: Layer = (2, 0)
    SLAB90: Layer = (3, 0)
    DEEPTRENCH: Layer = (4, 0)
    GE: Layer = (5, 0)
    WGN: Layer = (34, 0)
    WGN_CLAD: Layer = (36, 0)

    N: Layer = (20, 0)
    NP: Layer = (22, 0)
    NPP: Layer = (24, 0)
    P: Layer = (21, 0)
    PP: Layer = (23, 0)
    PPP: Layer = (25, 0)
    GEN: Layer = (26, 0)
    GEP: Layer = (27, 0)

    HEATER: Layer = (47, 0)
    M1: Layer = (41, 0)
    M2: Layer = (45, 0)
    M3: Layer = (49, 0)
    VIAC: Layer = (40, 0)
    VIA1: Layer = (44, 0)
    VIA2: Layer = (43, 0)
    PADOPEN: Layer = (46, 0)

    DICING: Layer = (100, 0)
    NO_TILE_SI: Layer = (71, 0)
    PADDING: Layer = (67, 0)
    DEVREC: Layer = (68, 0)
    FLOORPLAN: Layer = (64, 0)
    TEXT: Layer = (66, 0)
    PORT: Layer = (1, 10)
    PORTE: Layer = (1, 11)
    PORTH: Layer = (70, 0)
    SHOW_PORTS: Layer = (1, 12)
    LABEL: Layer = (201, 0)
    LABEL_SETTINGS: Layer = (202, 0)
    TE: Layer = (203, 0)
    TM: Layer = (204, 0)
    DRC_MARKER: Layer = (205, 0)
    LABEL_INSTANCE: Layer = (206, 0)
    ERROR_MARKER: Layer = (207, 0)
    ERROR_PATH: Layer = (208, 0)


LAYER = LayerMap()


def get_layer_stack(
    thickness_wg: float = 220 * nm,
    thickness_clad: float = 3.0,
    thickness_nitride: float = 350 * nm,
    gap_silicon_to_nitride: float = 100 * nm,
) -> LayerStack:
    """Returns generic LayerStack.

    based on paper https://www.degruyter.com/document/doi/10.1515/nanoph-2013-0034/html

    Args:
        thickness_wg: waveguide thickness.
        thickness_clad: cladding.
        thickness_nitride: for nitride.
        gap_silicon_to_nitride: in um.
    """
    return LayerStack(
        layers=dict(
            core=LayerLevel(
                layer=LAYER.WG,
                thickness=thickness_wg,
                zmin=0.0,
                material="si",
            ),
            clad=LayerLevel(
                layer=LAYER.WGCLAD,
                zmin=0.0,
                material="sio2",
                thickness=thickness_clad,
            ),
            slab150=LayerLevel(
                layer=LAYER.SLAB150,
                thickness=150e-3,
                zmin=0,
                material="si",
            ),
            slab90=LayerLevel(
                layer=LAYER.SLAB90,
                thickness=90e-3,
                zmin=0.0,
                material="si",
            ),
            nitride=LayerLevel(
                layer=LAYER.WGN,
                thickness=thickness_nitride,
                zmin=thickness_wg + gap_silicon_to_nitride,
                material="sin",
            ),
            ge=LayerLevel(
                layer=LAYER.GE,
                thickness=500e-3,
                zmin=thickness_wg,
                material="ge",
            ),
            via_contact=LayerLevel(
                layer=LAYER.VIAC,
                thickness=1100e-3,
                zmin=90e-3,
                material="Aluminum",
            ),
            metal1=LayerLevel(
                layer=LAYER.M1,
                thickness=750e-3,
                zmin=thickness_wg + 1100e-3,
                material="Aluminum",
            ),
            heater=LayerLevel(
                layer=LAYER.HEATER,
                thickness=750e-3,
                zmin=thickness_wg + 1100e-3,
                material="TiN",
            ),
            viac=LayerLevel(
                layer=LAYER.VIA1,
                thickness=1500e-3,
                zmin=thickness_wg + 1100e-3 + 750e-3,
                material="Aluminum",
            ),
            metal2=LayerLevel(
                layer=LAYER.M2,
                thickness=2000e-3,
                zmin=thickness_wg + 1100e-3 + 750e-3 + 1.5,
                material="Aluminum",
            ),
        )
    )


LAYER_STACK = get_layer_stack()


if __name__ == "__main__":
    ls = LAYER_STACK
    ls.get_klayout_3d_script()



from __future__ import annotations

import numpy as np
import pytest

from gdsfactory.simulation.simphony.components import (
    circuit_factory,
    circuit_names,
    component_names,
    model_factory,
)
from gdsfactory.simulation.simphony.get_transmission import get_transmission


@pytest.mark.parametrize("component_type", component_names)
def test_elements(component_type, data_regression) -> None:
    c = model_factory[component_type]()
    wav = np.linspace(1520, 1570, 3) * 1e-9
    f = 3e8 / wav
    s = c.s_parameters(f)
    _, rows, cols = np.shape(s)
    sdict = {
        f"s{i+1}{j+1}": np.round(np.abs(s[:, i, j]), decimals=3).tolist()
        for i in range(rows)
        for j in range(cols)
    }
    data_regression.check(sdict)


@pytest.mark.parametrize("component_type", circuit_names)
def test_circuits(component_type, data_regression) -> None:
    c = circuit_factory[component_type]()
    r = get_transmission(c, pin_in=c.pins[0].name, pin_out=c.pins[-1].name, num=3)
    s = np.round(r["s"], decimals=3).tolist()
    data_regression.check(dict(w=r["wavelengths"].tolist(), s=s))



from __future__ import annotations

from typing import Optional, Tuple
from functools import partial

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.typings import Ints, LayerSpec


@cell
def rectangle(
    size: Tuple[float, float] = (4.0, 2.0),
    layer: LayerSpec = "WG",
    centered: bool = False,
    port_type: Optional[str] = "placement",
    port_orientations: Optional[Ints] = (180, 90, 0, -90),
) -> Component:
    """Returns a rectangle.

    Args:
        size: (tuple) Width and height of rectangle.
        layer: Specific layer to put polygon geometry on.
        centered: True sets center to (0, 0), False sets south-west to (0, 0).
        port_type: optical, electrical.
        port_orientations: list of port_orientations to add.
    """
    c = Component()
    ref = c << compass(
        size=size, layer=layer, port_type=port_type, port_orientations=port_orientations
    )
    if not centered:
        ref.move((size[0] / 2, size[1] / 2))
    if port_type:
        c.add_ports(ref.ports)
    return c


fiber_size = 10.4
marker_te = partial(rectangle, size=[fiber_size, fiber_size], layer="TE", centered=True)
marker_tm = partial(rectangle, size=[fiber_size, fiber_size], layer="TM", centered=True)


if __name__ == "__main__":
    c = rectangle(size=(3, 2), centered=False, layer=(2, 3))
    # c = rectangle(size=(3, 2), centered=True, layer=(2, 3))
    print(c.ports)
    print(c.name)
    c.show(show_ports=True)



from __future__ import annotations

from gdsfactory.simulation.gmsh.mesh import create_physical_mesh, mesh_from_polygons
from gdsfactory.simulation.gmsh.meshtracker import MeshTracker
from gdsfactory.simulation.gmsh.parse_gds import (
    cleanup_component,
    fuse_polygons,
    round_coordinates,
    tile_shapes,
    to_polygons,
)
from gdsfactory.simulation.gmsh.parse_layerstack import (
    get_layer_overlaps_z,
    get_layers_at_z,
    list_unique_layerstack_z,
    map_unique_layerstack_z,
    order_layerstack,
)
from gdsfactory.simulation.gmsh.uz_xsection_mesh import (
    get_u_bounds_layers,
    get_u_bounds_polygons,
    get_uz_bounds_layers,
    uz_xsection_mesh,
)
from gdsfactory.simulation.gmsh.xy_xsection_mesh import xy_xsection_mesh

__all__ = [
    "mesh_from_polygons",
    "create_physical_mesh",
    "uz_xsection_mesh",
    "xy_xsection_mesh",
    "get_uz_bounds_layers",
    "get_u_bounds_layers",
    "get_u_bounds_polygons",
    "MeshTracker",
    "cleanup_component",
    "fuse_polygons",
    "round_coordinates",
    "to_polygons",
    "tile_shapes",
    "get_layers_at_z",
    "order_layerstack",
    "list_unique_layerstack_z",
    "map_unique_layerstack_z",
    "get_layer_overlaps_z",
]
__version__ = "0.0.2"



"""Install Klayout and GIT plugins."""
from __future__ import annotations

import configparser
import os
import pathlib
import shutil
import sys
from typing import Optional


def remove_path_or_dir(dest: pathlib.Path):
    if dest.is_dir():
        os.unlink(dest)
    else:
        os.remove(dest)


def make_link(src, dest, overwrite: bool = True) -> None:
    dest = pathlib.Path(dest)
    if dest.exists() and not overwrite:
        print(f"{dest} already exists")
        return
    if dest.exists() or dest.is_symlink():
        print(f"removing {dest} already installed")
        remove_path_or_dir(dest)
    try:
        os.symlink(src, dest, target_is_directory=True)
    except OSError as err:
        print("Could not create symlink!")
        print("     Error: ", err)
        if sys.platform == "win32":
            # https://stackoverflow.com/questions/32877260/privlege-error-trying-to-create-symlink-using-python-on-windows-10
            shutil.copy(src, dest)
    print("Symlink made:")
    print(f"From: {src}")
    print(f"To:   {dest}")


def install_gdsdiff() -> None:
    """Install gdsdiff tool."""
    home = pathlib.Path.home()
    git_config_path = home / ".gitconfig"
    git_attributes_path = home / ".gitattributes"

    if git_config_path.exists():
        git_config_str = open(git_config_path).read()
    else:
        git_config_str = "empty"

    git_attributes_str = (
        open(git_attributes_path).read() if git_attributes_path.exists() else "empty"
    )

    if "gds_diff" not in git_config_str:
        write_git_config(git_config_path)
    if "gds_diff" not in git_attributes_str:
        print("Appending the gdsdiff command to your ~/.gitattributes")

        with open(git_attributes_path, "a") as f:
            f.write("*.gds diff=gds_diff\n")


def write_git_config(git_config_path):
    """Write GIT config."""
    print("gdsdiff shows boolean differences in Klayout")
    print("git diff FILE.GDS")
    print("Appending the gdsdiff command to your ~/.gitconfig")

    config = configparser.RawConfigParser()
    config.read(git_config_path)
    key = 'diff "gds_diff"'

    if key not in config.sections():
        config.add_section(key)
        config.set(key, "command", "python -m gdsfactory.gdsdiff.gds_diff_git")
        config.set(key, "binary", "True")

        with open(git_config_path, "w+") as f:
            config.write(f, space_around_delimiters=True)


def get_klayout_path() -> pathlib.Path:
    """Returns KLayout path."""
    klayout_folder = "KLayout" if sys.platform == "win32" else ".klayout"
    home = pathlib.Path.home()
    return home / klayout_folder


def copy(src: pathlib.Path, dest: pathlib.Path) -> None:
    """Copy overwriting file or directory."""
    dest_folder = dest.parent
    dest_folder.mkdir(exist_ok=True, parents=True)

    if dest.exists() or dest.is_symlink():
        print(f"removing {dest} already installed")
        remove_path_or_dir(dest)

    if src.is_dir():
        shutil.copytree(src, dest)
    else:
        shutil.copy(src, dest)
    print(f"{src} copied to {dest}")


def _install_to_klayout(
    src: pathlib.Path, klayout_subdir_name: str, package_name: str
) -> None:
    """Install into KLayout technology.

    Equivalent to using KLayout package manager.

    """
    klayout_folder = "KLayout" if sys.platform == "win32" else ".klayout"
    subdir = pathlib.Path.home() / klayout_folder / klayout_subdir_name
    dest = subdir / package_name
    subdir.mkdir(exist_ok=True, parents=True)
    make_link(src, dest)


def install_klayout_package() -> None:
    """Install gdsfactory KLayout package.

    Equivalent to using KLayout package manager.

    """
    cwd = pathlib.Path(__file__).resolve().parent
    _install_to_klayout(
        src=cwd / "generic_tech" / "klayout",
        klayout_subdir_name="salt",
        package_name="gdsfactory",
    )


def install_klayout_technology(tech_dir: pathlib.Path, tech_name: Optional[str] = None):
    """Install technology to KLayout."""
    _install_to_klayout(
        src=tech_dir,
        klayout_subdir_name="tech",
        package_name=tech_name or tech_dir.name,
    )


if __name__ == "__main__":
    cwd = pathlib.Path(__file__).resolve().parent
    home = pathlib.Path.home()
    src = cwd / "generic_tech" / "klayout" / "tech"

    install_gdsdiff()
    install_klayout_package()



"""pack a list of components into a grid.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""
from __future__ import annotations

from typing import Optional, Tuple

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.component_layout import Group
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.components.triangles import triangle
from gdsfactory.typings import Anchor, ComponentSpec, Float2


@cell
def grid(
    components: Optional[Tuple[ComponentSpec, ...]] = None,
    spacing: Tuple[float, float] = (5.0, 5.0),
    separation: bool = True,
    shape: Optional[Tuple[int, int]] = None,
    align_x: str = "x",
    align_y: str = "y",
    edge_x: str = "x",
    edge_y: str = "ymax",
    rotation: int = 0,
    h_mirror: bool = False,
    v_mirror: bool = False,
    add_ports_prefix: bool = True,
    add_ports_suffix: bool = False,
) -> Component:
    """Returns Component with a 1D or 2D grid of components.

    Args:
        components: Iterable to be placed onto a grid. (can be 1D or 2D).
        spacing: between adjacent elements on the grid, can be a tuple for
            different distances in height and width.
        separation: If True, guarantees elements are separated with fixed spacing
            if False, elements are spaced evenly along a grid.
        shape: x, y shape of the grid (see np.reshape).
            If no shape and the list is 1D, if np.reshape were run with (1, -1).
        align_x: {'x', 'xmin', 'xmax'} for x (column) alignment along.
        align_y: {'y', 'ymin', 'ymax'} for y (row) alignment along.
        edge_x: {'x', 'xmin', 'xmax'} for x (column) (ignored if separation = True).
        edge_y: {'y', 'ymin', 'ymax'} for y (row) along (ignored if separation = True).
        rotation: for each component in degrees.
        h_mirror: horizontal mirror y axis (x, 1) (1, 0). most common mirror.
        v_mirror: vertical mirror using x axis (1, y) (0, y).
        add_ports_prefix: adds port names with prefix.
        add_ports_suffix: adds port names with suffix.

    Returns:
        Component containing components grid.

    .. plot::
        :include-source:

        import gdsfactory as gf

        components = [gf.components.triangle(x=i) for i in range(1, 10)]
        c = gf.grid(
            components,
            shape=(1, len(components)),
            rotation=0,
            h_mirror=False,
            v_mirror=True,
            spacing=(100, 100),
        )
        c.plot()

    """
    components = components or [triangle(x=i) for i in range(1, 10)]
    device_array = np.asarray(components)

    # Check arguments
    if device_array.ndim not in (1, 2):
        raise ValueError("grid() The components needs to be 1D or 2D.")
    if shape is not None and len(shape) != 2:
        raise ValueError(
            "grid() shape argument must be None or"
            f" have a length of 2, for example shape=(4,6), got {shape}"
        )

    # Check that shape is valid and reshape array if needed
    if (shape is None) and (device_array.ndim == 2):  # Already in desired shape
        shape = device_array.shape
    elif (shape is None) and (device_array.ndim == 1):
        shape = (device_array.size, -1)
    elif 0 < shape[0] * shape[1] < device_array.size:
        raise ValueError(
            f"Shape {shape} is too small for all {device_array.size} components"
        )
    else:
        if np.min(shape) == -1:
            remainder = np.max(shape) - device_array.size % np.max(shape)
        else:
            remainder = shape[0] * shape[1] - device_array.size
        if remainder != 0:
            device_array = np.append(
                device_array,
                [
                    None,
                ]
                * remainder,
            )
    device_array = np.reshape(device_array, shape)
    prefix_to_ref = {}

    D = Component()
    ref_array = np.empty(device_array.shape, dtype=object)
    dummy = Component("dummy")
    for idx, d in np.ndenumerate(device_array):
        if d is not None:
            d = d() if callable(d) else d
            ref = D.add_ref(d, rotation=rotation, x_reflection=h_mirror)
            if v_mirror:
                ref.mirror_y()
            ref_array[idx] = ref
            prefix = f"{idx}"
            prefix = prefix.replace(" ", "")
            prefix = prefix.replace(",", "_")
            prefix = prefix.replace("(", "")
            prefix = prefix.replace(")", "")
            ref.name = prefix
            prefix_to_ref[prefix] = ref

        else:
            ref_array[idx] = D << dummy  # Create dummy devices

    rows = [Group(ref_array[n, :]) for n in range(ref_array.shape[0])]
    cols = [Group(ref_array[:, n]) for n in range(ref_array.shape[1])]

    # Align rows and columns independently
    for r in rows:
        r.align(alignment=align_y)
    for c in cols:
        c.align(alignment=align_x)

    # Distribute rows and columns
    Group(cols).distribute(
        direction="x", spacing=spacing[0], separation=separation, edge=edge_x
    )
    Group(rows[::-1]).distribute(
        direction="y", spacing=spacing[1], separation=separation, edge=edge_y
    )

    for prefix, ref in prefix_to_ref.items():
        if add_ports_prefix:
            D.add_ports(ref.ports, prefix=f"{prefix}_")
        elif add_ports_suffix:
            D.add_ports(ref.ports, suffix=f"_{prefix}")
        else:
            D.add_ports(ref.ports)

    return D


@cell
def grid_with_text(
    components: Optional[Tuple[ComponentSpec, ...]] = None,
    text_prefix: str = "",
    text_offsets: Tuple[Float2, ...] = ((0, 0),),
    text_anchors: Tuple[Anchor, ...] = ("cc",),
    text: Optional[ComponentSpec] = text_rectangular,
    labels: Optional[Tuple[str, ...]] = None,
    **kwargs,
) -> Component:
    """Returns Component with 1D or 2D grid of components with text labels.

    Args:
        components: Iterable to be placed onto a grid. (can be 1D or 2D).
        text_prefix: for labels. For example. 'A' will produce 'A1', 'A2', ...
        text_offsets: relative to component anchor. Defaults to center.
        text_anchors: relative to component (ce cw nc ne nw sc se sw center cc).
        text: function to add text labels.
        labels: optional, specify a tuple of labels rather than using a text_prefix.

    keyword Args:
        spacing: between adjacent elements on the grid, can be a tuple for
          different distances in height and width.
        separation: If True, guarantees elements are separated with fixed spacing
          if False, elements are spaced evenly along a grid.
        shape: x, y shape of the grid (see np.reshape).
          If no shape and the list is 1D, if np.reshape were run with (1, -1).
        align_x: {'x', 'xmin', 'xmax'} to perform the x (column) alignment along.
        align_y: {'y', 'ymin', 'ymax'} to perform the y (row) alignment along.
        edge_x: {'x', 'xmin', 'xmax'} to perform the x (column) distribution
            ignored if separation = True.
        edge_y: {'y', 'ymin', 'ymax'} to perform the y (row) distribution along
            ignored if separation = True.
        rotation: for each reference in degrees.


    .. plot::
        :include-source:

        import gdsfactory as gf

        components = [gf.components.triangle(x=i) for i in range(1, 10)]
        c = gf.grid_with_text(
            components,
            shape=(1, len(components)),
            rotation=0,
            h_mirror=False,
            v_mirror=True,
            spacing=(100, 100),
            text_offsets=((0, 100), (0, -100)),
            text_anchors=("nc", "sc"),
        )
        c.plot()

    """
    c = Component()
    g = grid(components=components, **kwargs)
    c << g
    if text:
        for i, ref in enumerate(g.named_references.values()):
            for text_offset, text_anchor in zip(text_offsets, text_anchors):
                if labels:
                    if len(labels) > i:
                        label = labels[i]
                    # skip labels for dummy components
                    else:
                        continue
                else:
                    label = f"{text_prefix}{i}"
                t = c << text(label)
                t.move(np.array(text_offset) + getattr(ref.size_info, text_anchor))
    return c


def test_grid() -> Component:
    import gdsfactory as gf

    c = [gf.components.straight(length=i) for i in [1, 1, 2]]
    c = grid(
        c,
        shape=(2, 2),
        rotation=0,
        h_mirror=False,
        v_mirror=False,
        spacing=(10, 10),
    )
    # assert np.isclose(c.ports["1_1_o1"].center[0], 13.0), c.ports["1_1_o1"].center[0]
    return c


if __name__ == "__main__":
    c = test_grid()
    # import gdsfactory as gf

    # components = [gf.components.rectangle(size=(i, i)) for i in range(40, 66, 5)]
    # components = [gf.components.rectangle(size=(i, i)) for i in range(40, 66, 5)]
    # c = [gf.components.triangle(x=i) for i in range(1, 10)]
    # print(len(c))

    # c = [gf.components.straight(length=i) for i in range(1, 5)]
    # c = grid(
    #     c,
    #     shape=(2, 2),
    #     rotation=0,
    #     h_mirror=False,
    #     v_mirror=False,
    #     spacing=(10, 10),
    #     # text_offsets=((0, 100), (0, -100)),
    #     # text_anchors=("nc", "sc"),
    # )
    c.show(show_ports=True)



import numpy as np
import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec
from gdsfactory.components.bend_s import bend_s


@gf.cell
def mmi2x2_with_sbend(
    with_sbend: bool = True,
    s_bend: ComponentSpec = bend_s,
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    """Returns mmi2x2 for Cband.

    C_band 2x2MMI in 220nm thick silicon
    https://opg.optica.org/oe/fulltext.cfm?uri=oe-25-23-28957&id=376719

    Args:
        with_sbend: add sbend.
        s_bend: S-bend spec.
        cross_section: spec.
    """

    def mmi_widths(t):
        # Note: Custom width/offset functions MUST be vectorizable--you must be able
        # to call them with an array input like my_custom_width_fun([0, 0.1, 0.2, 0.3, 0.4])
        widths = np.array([2 * 0.7 + 0.2, 1.48, 1.48, 1.48, 1.6])
        return widths

    c = gf.Component()

    P = gf.path.straight(length=2 * 2.4 + 2 * 1.6, npoints=5)

    xs = gf.get_cross_section(cross_section)
    xs.width = mmi_widths
    c << gf.path.extrude(P, cross_section=xs)

    # Add input and output tapers
    top_input_block = c << gf.components.taper(
        length=1, width1=0.5, width2=0.7, cross_section=cross_section
    ).move((-1, 0.45))
    bottom_input_block = c << gf.components.taper(
        length=1, width1=0.5, width2=0.7, cross_section=cross_section
    ).move((-1, -0.45))

    top_output_block = c << gf.components.taper(
        length=1, width1=0.7, width2=0.5, cross_section=cross_section
    ).move((8, 0.45))
    bottom_output_block = c << gf.components.taper(
        length=1, width1=0.7, width2=0.5, cross_section=cross_section
    ).move((8, -0.45))

    if with_sbend:
        sbend = gf.get_component(s_bend, cross_section=cross_section)

        top_input_sbend_ref = c << sbend
        top_input_sbend_ref.mirror([0, 1])
        bottom_input_sbend_ref = c << sbend
        top_output_sbend_ref = c << sbend
        bottom_output_sbend_ref = c << sbend
        bottom_output_sbend_ref.mirror([0, 1])

        top_input_sbend_ref.connect("o1", destination=top_input_block.ports["o1"])
        bottom_input_sbend_ref.connect("o1", destination=bottom_input_block.ports["o1"])
        top_output_sbend_ref.connect("o1", destination=top_output_block.ports["o2"])
        bottom_output_sbend_ref.connect(
            "o1", destination=bottom_output_block.ports["o2"]
        )

        c.add_port("o1", port=bottom_input_sbend_ref.ports["o2"])
        c.add_port("o2", port=top_input_sbend_ref.ports["o2"])
        c.add_port("o3", port=top_output_sbend_ref.ports["o2"])
        c.add_port("o4", port=bottom_output_sbend_ref.ports["o2"])

    else:
        c.add_port("o2", port=top_input_block.ports["o1"])
        c.add_port("o1", port=bottom_input_block.ports["o1"])
        c.add_port("o3", port=top_output_block.ports["o2"])
        c.add_port("o4", port=bottom_output_block.ports["o2"])
    return c


if __name__ == "__main__":
    c = mmi2x2_with_sbend(
        with_sbend=True,
        cross_section=dict(cross_section="strip", settings=dict(layer=(2, 0))),
    )
    # c = mmi2x2_with_sbend(with_sbend=False)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler_ring import coupler_ring as coupler_ring_function
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentSpec


@gf.cell
def ring_single(
    gap: float = 0.2,
    radius: float = 10.0,
    length_x: float = 4.0,
    length_y: float = 0.6,
    coupler_ring: ComponentSpec = coupler_ring_function,
    bend: ComponentSpec = bend_euler,
    cross_section: ComponentSpec = "strip",
    **kwargs,
) -> gf.Component:
    """Returns a single ring.

    ring coupler (cb: bottom) connects to two vertical straights (sl: left, sr: right),
    two bends (bl, br) and horizontal straight (wg: top)

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler_ring: ring coupler spec.
        bend: 90 degrees bend spec.
        cross_section: cross_section spec.
        kwargs: cross_section settings.

    .. code::

          bl-st-br
          |      |
          sl     sr length_y
          |      |
         --==cb==-- gap

          length_x
    """
    gap = gf.snap.snap_to_grid(gap, nm=2)

    c = gf.Component()
    cb = c << gf.get_component(
        coupler_ring,
        bend=bend,
        gap=gap,
        radius=radius,
        length_x=length_x,
        cross_section=cross_section,
        **kwargs,
    )
    sy = straight(length=length_y, cross_section=cross_section, **kwargs)
    b = gf.get_component(bend, cross_section=cross_section, radius=radius, **kwargs)
    sx = straight(length=length_x, cross_section=cross_section, **kwargs)
    sl = c << sy
    sr = c << sy
    bl = c << b
    br = c << b
    st = c << sx

    sl.connect(port="o1", destination=cb.ports["o2"])
    bl.connect(port="o2", destination=sl.ports["o2"])

    st.connect(port="o2", destination=bl.ports["o1"])
    br.connect(port="o2", destination=st.ports["o1"])
    sr.connect(port="o1", destination=br.ports["o1"])
    sr.connect(port="o2", destination=cb.ports["o3"])

    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o1", port=cb.ports["o1"])
    return c


if __name__ == "__main__":
    # c = ring_single(layer=(2, 0), cross_section_factory=gf.cross_section.pin, width=1)
    # c = ring_single(width=2, gap=1, layer=(2, 0), radius=7, length_y=1)
    # print(c.ports)

    # c = gf.routing.add_fiber_array(ring_single)
    c = ring_single(cross_section="rib", width=2)
    c.show(show_ports=True)

    # cc = gf.add_pins(c)
    # print(c.settings)
    # print(c.settings)
    # cc.show(show_ports=True)



import gmsh


def global_callback_refinement(
    model,
    global_meshsize_array,
    global_meshsize_interpolant_func,
):
    """Refine gmsh/pygmsh model according to a global callback dict.

    Args:
        model
        global_meshsize_array
        global_meshsize_interpolant_func
    Returns:
        model
    """
    global_meshsize_interpolant = global_meshsize_interpolant_func(
        global_meshsize_array[:, 0:3], global_meshsize_array[:, 3]
    )

    def meshSizeCallback(dim, tag, x, y, z, lc):
        return min(lc, global_meshsize_interpolant(x, y, z))

    gmsh.model.mesh.setSizeCallback(meshSizeCallback)

    # Turn off default meshing options
    gmsh.model.mesh.MeshSizeFromPoints = 0
    gmsh.model.mesh.MeshSizeFromCurvature = 0
    gmsh.model.mesh.MeshSizeExtendFromBoundary = 0

    return model


def surface_interface_refinement(
    model,
    meshtracker,
    resolutions,
    default_resolution_min,
    default_resolution_max,
):
    """Refine gmsh/pygmsh model according to a resolutions dict.

    Args:
        model
        meshtracker
        resolutions
        default_resolution_min
        default_resolution_max
    Returns:
        model
        meshtracker
    """
    n = 0
    refinement_fields = []
    for label, resolution in resolutions.items():
        # Inside surface
        mesh_resolution = resolution["resolution"]
        gmsh.model.mesh.field.add("MathEval", n)
        gmsh.model.mesh.field.setString(n, "F", f"{mesh_resolution}")
        gmsh.model.mesh.field.add("Restrict", n + 1)
        gmsh.model.mesh.field.setNumber(n + 1, "InField", n)
        gmsh.model.mesh.field.setNumbers(
            n + 1,
            "SurfacesList",
            meshtracker.get_gmsh_xy_surfaces_from_label(label),
        )
        # Around surface
        mesh_distance = resolution["distance"]
        gmsh.model.mesh.field.add("Distance", n + 2)
        gmsh.model.mesh.field.setNumbers(
            n + 2, "CurvesList", meshtracker.get_gmsh_xy_lines_from_label(label)
        )
        gmsh.model.mesh.field.setNumber(n + 2, "Sampling", 100)
        gmsh.model.mesh.field.add("Threshold", n + 3)
        gmsh.model.mesh.field.setNumber(n + 3, "InField", n + 2)
        gmsh.model.mesh.field.setNumber(n + 3, "SizeMin", mesh_resolution)
        gmsh.model.mesh.field.setNumber(n + 3, "SizeMax", default_resolution_max)
        gmsh.model.mesh.field.setNumber(n + 3, "DistMin", 0)
        gmsh.model.mesh.field.setNumber(n + 3, "DistMax", mesh_distance)
        refinement_fields.extend((n + 1, n + 3))
        n += 4

    # Use the smallest element size overall
    gmsh.model.mesh.field.add("Min", n)
    gmsh.model.mesh.field.setNumbers(n, "FieldsList", refinement_fields)
    gmsh.model.mesh.field.setAsBackgroundMesh(n)

    # Turn off default meshing options
    gmsh.model.mesh.MeshSizeFromPoints = 0
    gmsh.model.mesh.MeshSizeFromCurvature = 0
    gmsh.model.mesh.MeshSizeExtendFromBoundary = 0

    return model, meshtracker
